# (C) Copyright 2022 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

import json
import logging
import re

from pyaoscx.exceptions.generic_op_error import GenericOperationError
from pyaoscx.exceptions.response_error import ResponseError

from pyaoscx.utils import util as utils

from pyaoscx.pyaoscx_module import PyaoscxModule


class TunnelEndpoint(PyaoscxModule):
    """
    Provide configuration for Tunnel Endpoints (TEP) on AOS-CX devices.
    """

    collection_uri = "system/interfaces/{intf_name}/tunnel_endpoints"
    object_uri = collection_uri + "/{vrf_name},{origin},{destination}"

    indices = ["vrf_name", "origin", "destination"]
    resource_uri_name = "tunnel_endpoints"

    def __init__(
        self,
        session,
        interface,
        network_id,
        destination,
        origin="static",
        vrf=None,
        **kwargs
    ):

        self.session = session
        if vrf is None:
            vrf = session.api.get_module(session, "Vrf", "default")
        self.vrf = vrf
        self.vrf_name = vrf.name
        self.network_id = network_id
        self.destination = destination
        self.origin = origin
        self.interface = interface
        self.tep_id = "{0},{1},{2}".format(
            self.vrf_name, self.origin, self.destination
        )
        self.config_attrs = []
        self.materialized = False
        self._original_attributes = {}
        utils.set_creation_attrs(self, **kwargs)
        uri_indices = {
            "intf_name": self.interface.percents_name,
            "vrf_name": self.vrf_name,
            "origin": self.origin,
            "destination": self.destination,
        }
        self.base_uri = self.collection_uri.format(**uri_indices)
        self.path = self.object_uri.format(**uri_indices)
        self.__modified = False

    def get(self, depth=None, selector=None):
        """
        Perform a GET call to retrieve data for a TEP table entry and fill the
            object with the incoming attributes.

        :param depth: Integer deciding how many levels into the API JSON that
            references will be returned.
        :param selector: Alphanumeric option to select specific information to
            return.
        :return: Returns True if no exception is raised.
        """
        logging.info("Retrieving %s from switch", self)
        # this is common for all PyaoscxModule derived classes
        self._get_and_copy_data(depth, selector, self.indices)
        self.materialized = True
        vni_urls = self.network_id
        vni_url = vni_urls[next(iter(vni_urls))]
        Vni = self.session.api.get_module_class(self.session, "Vni")
        _, vni = Vni.from_uri(self.session, self.interface, vni_url)
        self.network_id = vni
        return True

    @classmethod
    def get_all(cls, session, parent_interface):
        """
        Perform a GET call to retrieve all Tunnel Endpoints and create a
            dictionary containing each respective TEP.

        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param parent_interface: Interface attached to this tunnel
        :return: Dictionary containing TEP IDs as keys and a TEP object as
            value.
        """
        logging.info("Retrieving all %s data from switch", cls.__name__)
        uri_indices = {"intf_name": parent_interface.percents_name}
        uri = cls.collection_uri.format(**uri_indices)
        try:
            response = session.request("GET", uri)
        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        tep_collection = {}
        data = json.loads(response.text)
        for uri in data.values():
            tep_id, tep = cls.from_uri(session, uri)
            tep_collection[tep_id] = tep

        return tep_collection

    @PyaoscxModule.connected
    def apply(self):
        """
        Main method used to either create a TEP, or update an existing one.
            Checks whether the TEP exists in the switch. Calls self.update()
            if TEP is being updated. Calls self.create() if the TEP doesn't
            exist in the switch.

        :return: Boolean, True if object was created or modified.
        """
        if self.materialized:
            return self.update()
        return self.create()

    @PyaoscxModule.connected
    def update(self):
        """
        Perform a PUT call to apply changes to an existing TEP table entry.

        :return: True if Object was modified and a PUT request was
            made.
        """
        tep_data = {}
        tep_data["network_id"] = self.network_id.get_info_format()
        self.__modified = self._put_data(tep_data)
        return self.__modified

    @PyaoscxModule.connected
    def create(self):
        """
        Perform a POST call to create a new TEP using the object's attributes
            as POST body. Exception is raised if object is unable to be
            created.

        :return: Boolean, True if TEP was created.
        """
        tep_data = {}
        tep_data["destination"] = self.destination
        tep_data["interface"] = self.interface.get_info_format()
        tep_data["network_id"] = self.network_id.get_info_format()
        tep_data["origin"] = self.origin
        tep_data["vrf"] = self.vrf.get_info_format()

        self.__modified = self._post_data(tep_data)
        return self.__modified

    @PyaoscxModule.connected
    def delete(self):
        self._send_data(self.path, None, "DELETE", "Delete")
        # Delete object attributes
        utils.delete_attrs(self, self.config_attrs)

    @classmethod
    def from_response(cls, session, response_data):
        """
        Create a TEP object given a response_dat
        """
        if isinstance(response_data, dict):
            uri = list(response_data.values())[0]
        else:
            uri = response_data

        return cls.from_uri(session, uri)

    def __str__(self):
        tep_str = "Tunnel {0},{1},{2} for interface {3}".format(
            self.vrf.name, self.origin, self.destination, self.interface.name
        )
        return tep_str

    @classmethod
    def from_uri(cls, session, uri):
        """
        Create a TEP object given a TEP URI.
        """
        # Obtain ID from URI like:
        # system/interfaces/{intf_name}/tunnel_endpoints/{vrf_name},{origin},{destination}
        index_pattern = re.compile(
            r"(.*)interfaces/(?P<intf>.+)/tunnel_endpoints/(?P<vrf>.+),(?P<origin>.+),(?P<dest>.+)"  # NOQA
        )
        pattern_match = index_pattern.match(uri)
        intf_name = pattern_match.group("intf")
        vrf_name = pattern_match.group("vrf")
        origin = pattern_match.group("origin")
        destination = pattern_match.group("dest")

        interface = session.api.get_module(session, "Interface", intf_name)
        vrf = session.api.get_module(session, "Vrf", vrf_name)
        tep = cls(
            session,
            interface=interface,
            network_id=None,
            destination=destination,
            origin=origin,
            vrf=vrf,
        )

        return tep.tep_id, tep

    @PyaoscxModule.deprecated
    def get_uri(self):
        """
        Method used to obtain the specific TEP URI.

        :return: Object's URI.
        """
        uri = "{0}{1}".format(self.session.resource_prefix, self.path)
        return uri

    @PyaoscxModule.deprecated
    def get_info_format(self):
        """
        Method used to obtain correct object format for referencing inside
            other objects.

        :return: Object format depending on the API Version.
        """
        return self.session.api.get_index(self)

    @property
    def modified(self):
        """
        Return boolean with whether this object has been modified.
        """
        return self.__modified

    @PyaoscxModule.deprecated
    def was_modified(self):
        """
        Getter method for the __modified attribute.

        :return: Boolean True if the object was recently modified.
        """
        return self.modified



# (C) Copyright 2021-2023 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

import json
import logging

from pyaoscx.exceptions.generic_op_error import GenericOperationError
from pyaoscx.exceptions.response_error import ResponseError

from pyaoscx.utils import util as utils

from pyaoscx.queue_profile import QueueProfile

from pyaoscx.pyaoscx_module import PyaoscxModule


class QueueProfileEntry(PyaoscxModule):
    """
    Provide configuration management for Queue Profile Entries on AOS-CX
        devices.
    """

    collection_uri = "system/q_profiles/{name}/q_profile_entries"
    object_uri = collection_uri + "/{queue_number}"
    resource_name = "queue_number"
    indices = ["queue_number"]

    def __init__(self, session, queue_number, parent_profile, **kwargs):
        self.__profile = parent_profile
        self.__queue_number = queue_number
        self.session = session
        # List used to determine attributes related to the
        # Queue profile configuration
        self.config_attrs = []
        self.materialized = False

        # Attribute dictionary used to manage the original data
        # obtained from the GET request
        self._original_attributes = {}
        utils.set_creation_attrs(self, **kwargs)

        # Used to know if the object was changed since the last
        # request
        self.__modified = False

        # Build the URI that identifies the current Queue profile
        self.path = self.object_uri.format(
            name=self.__profile.name, queue_number=queue_number
        )
        self.base_uri = self.collection_uri.format(name=self.__profile.name)

    @property
    def queue_number(self):
        return self.__queue_number

    @PyaoscxModule.connected
    def get(self, depth=None, selector=None):
        """
        Perform a GET call to retrieve data for a Queue Profile Entry and fill
            the object with the incoming attributes.

        :param depth: Integer deciding how many levels into the API JSON that
            references will be returned.
        :param selector: Alphanumeric option to select specific information
            to return.
        :return: Returns True if there is not an exception raised.
        """
        logging.info("Retrieving %s from switch", self)

        selector = selector or self.session.api.default_selector

        data = self._get_data(depth, selector)

        # Add dictionary as attributes for the object
        utils.create_attrs(self, data)

        # Update the original attributes
        self._original_attributes = data

        self.materialized = True
        return True

    @classmethod
    def get_all(cls, session, queue_profile_name):
        """
        Perform a GET call to retrieve all Queue Profile Entries  of the same
            profile and create a dictionary containing them.

        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param queue_profile_name: Name of the profile to which the entries
            belong.
        :return: Dictionary containing Queue Profile Entry names as keys
            and a Queue Profile Entry object as value.
        """
        logging.info("Retrieving all %s data from switch", cls.__name__)

        uri = cls.collection_uri.format(name=queue_profile_name)

        try:
            response = session.request("GET", uri)
        except Exception as exc:
            raise ResponseError("GET", exc)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        data = json.loads(response.text)

        entry_dict = {}
        # Get all URI elements in the form of a list
        uri_list = session.api.get_uri_from_data(data)

        for uri in uri_list:
            entry_number, entry = cls.from_uri(session, uri)
            entry_dict[entry_number] = entry

        return entry_dict

    @PyaoscxModule.connected
    def apply(self):
        """
        Main method used to either create or update an existing Queue
            Profile Entry. Checks whether the Queue Profile exists in the
            switch and calls self.update() or self.create() accordingly.

        :return modified: True if the object was modified.
        """
        if self.materialized:
            return self.update()
        else:
            return self.create()

    @PyaoscxModule.connected
    def update(self):
        """
        Perform a PUT call to apply changes to an existing Queue Profile Entry.

        :return: True if the object was modified and a PUT request was
            made.
        """
        data = utils.get_attrs(self, self.config_attrs)
        # Manually remove the queue_number
        if "queue_number" in data:
            del data["queue_number"]
        self.__modified = self._put_data(data)
        return self.__modified

    @PyaoscxModule.connected
    def create(self):
        """
        Perform a POST call to create a new Queue Profile Entry in the switch.

        :return: True if the object was modified.
        """
        data = utils.get_attrs(self, self.config_attrs)
        # Manually add the queue_number
        data["queue_number"] = self.queue_number
        self.__modified = self._post_data(data)
        return self.__modified

    @PyaoscxModule.connected
    def delete(self):
        """
        Perform a DELETE call to remove a Queue Profile Entry from the switch.
        """
        self._send_data(self.path, None, "DELETE", "Delete")
        utils.delete_attrs(self, self.config_attrs)

    @classmethod
    def from_uri(cls, session, uri):
        """
        Create a Queue profile object given an URI.

        :param cls: Object's class.
        :param session: Pyaoscx.Session objec used to represent a logical
            connection to the device.
        :param uri: a string with the URI.
        :return: tuple with the name and the Profile.
        """
        # Obtain the ID from URI
        # URI format is /system/q_profile/{name}/q_entry/{queue_number}
        parts = uri.split("/")
        profile_name = parts[-3]
        queue_number = parts[-1]
        profile = QueueProfile(session, profile_name)
        entry = cls(session, queue_number, profile)
        return queue_number, entry

    @classmethod
    def get_facts(cls, session, queue_profile_name):
        """
        Retrieve the information of all Queue profiles.

        :param cls: Class reference.
        :param session: Pyaoscx.Session object used to represent a logical
            connection to the device.
        :param queue_profile_name: Name of the profile to which the entries
            belong.
        :return: Dictionary containing the name as key and the facts as value.
        """
        logging.info("Retrieving Queue Profiles facts")

        depth = session.api.default_facts_depth

        uri = cls.collection_uri.format(name=queue_profile_name)

        try:
            response = session.request("GET", uri, params={"depth": depth})
        except Exception as exc:
            raise ResponseError("GET", exc)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        return json.loads(response.text)

    def __str__(self):
        return "Queue Profile Entry {0} of Profile {1}".format(
            self.queue_number, self.__profile.name
        )

    @property
    def modified(self):
        return self.__modified

    @PyaoscxModule.deprecated
    def was_modified(self):
        return self.modified



# (C) Copyright 2023 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

from __future__ import absolute_import, division, print_function

from ipaddress import IPv4Interface, IPv6Interface
from urllib.parse import quote_plus

from pyaoscx.exceptions.verification_error import VerificationError

from pyaoscx.device import Device


def overlapping_ips(
    session, interface, ip4_addr=None, ip6_addr=None, new_vrf="default"
):
    """
    Check if IPv4/IPv6 address is already configured.
    :param session: pyaoscx.Session object used to represent a logical
        connection to the device.
    :param interface: Interface to be configured.
    :param ip4_addr: IPv4 address to check.
    :param ip6_addr: IPv6 address to check.
    :param new_vrf: New VRF name for interface.
    """
    overlapping_ips = []

    device = Device(session)
    configuration = device.configuration()
    running_config = configuration.get_full_config()

    interfaces = running_config["Port"]
    interfaces.pop(quote_plus(interface), None)
    for intf in interfaces.values():
        old_vrf = intf["vrf"] if "vrf" in intf else "default"
        if new_vrf != old_vrf:
            continue
        if ip4_addr:
            new_net = IPv4Interface(ip4_addr).network
            if "ip4_address" in intf:
                old_net = IPv4Interface(intf["ip4_address"]).network
                if new_net.overlaps(old_net):
                    overlapping_ips.append(ip4_addr)
            if "ip4_address_secondary" in intf:
                ip4_secs = []
                if isinstance(intf["ip4_address_secondary"], str):
                    ip4_secs.append(intf["ip4_address_secondary"])
                elif isinstance(intf["ip4_address_secondary"], list):
                    ip4_secs.extend(intf["ip4_address_secondary"])
                for old_ip4 in ip4_secs:
                    old_net = IPv4Interface(old_ip4).network
                    if new_net.overlaps(old_net):
                        overlapping_ips.append(ip4_addr)
        if ip6_addr:
            new_net = IPv6Interface(ip6_addr).network
            if "ip6_addresses" in intf:
                for old_ip6 in intf["ip6_addresses"].keys():
                    old_net = IPv6Interface(old_ip6).network
                    print(old_net, new_net)
                    if new_net.overlaps(old_net):
                        overlapping_ips.append(ip6_addr)

    if overlapping_ips != []:
        raise VerificationError(
            "Following IPs already exist or are overlapping network "
            "in VRF {0}: {1}.".format(new_vrf, ",".join(overlapping_ips))
        )



# (C) Copyright 2019-2023 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

import json
import logging

from pyaoscx.exceptions.generic_op_error import GenericOperationError
from pyaoscx.exceptions.response_error import ResponseError
from pyaoscx.exceptions.verification_error import VerificationError

from pyaoscx.utils import util as utils
from pyaoscx.utils.list_attributes import ListDescriptor

from pyaoscx.aggregate_address import AggregateAddress
from pyaoscx.bgp_neighbor import BgpNeighbor

from pyaoscx.pyaoscx_module import PyaoscxModule


class BgpRouter(PyaoscxModule):
    """
    Provide configuration management for BGP Routers on AOS-CX devices.
    """

    indices = ["asn"]
    resource_uri_name = "bgp_routers"

    # Use to manage BGP Neighbors
    bgp_neighbors = ListDescriptor("bgp_neighbors")
    aggregate_addresses = ListDescriptor("aggregate_addresses")

    collection_uri = "system/vrfs/{name}/bgp_routers"
    object_uri = collection_uri + "/{asn}"

    def __init__(self, session, asn, parent_vrf, uri=None, **kwargs):
        self.session = session
        # Assign id
        self.__asn = asn
        self.__parent_vrf = parent_vrf
        self._uri = uri
        # List used to determine attributes related to the BGP configuration
        self.config_attrs = []
        self.materialized = False
        # Attribute dictionary used to manage the original data
        # obtained from the GET
        self._original_attributes = {}
        # Set arguments needed for correct creation
        utils.set_creation_attrs(self, **kwargs)
        # Use to manage BGP Neighbors
        self.bgp_neighbors = []
        # Use to manage Aggregate Addresses
        self.aggregate_addresses = []
        # Attribute used to know if object was changed recently
        self.__modified = False
        uri_indices = {
            "name": self.__parent_vrf.name,
            "asn": self.__asn,
        }
        self._uri_indices = uri_indices
        self.base_uri = self.collection_uri.format(**uri_indices)
        self.path = self.object_uri.format(**uri_indices)
        self.__parent_vrf.update_bgp_routers(self)

    @property
    def asn(self):
        return self.__asn

    @PyaoscxModule.connected
    def get(self, depth=None, selector=None):
        """
        Perform a GET call to retrieve data for a BGP Router table entry and
            fill the object with the incoming attributes

        :param depth: Integer deciding how many levels into the API JSON that
            references will be returned.
        :param selector: Alphanumeric option to select specific information to
            return.
        :return: Returns True if there is not an exception raised.
        """
        logging.info("Retrieving %s from switch", self)

        depth = depth or self.session.api.default_depth
        selector = selector or self.session.api.default_selector

        data = self._get_data(depth, selector)

        # Delete unwanted data
        if "bgp_neighbors" in data:
            data.pop("bgp_neighbors")
        if "aggregate_addresses" in data:
            data.pop("aggregate_addresses")

        # Add dictionary as attributes for the object
        utils.create_attrs(self, data)

        # Determines if the BGP Router is configurable
        if selector in self.session.api.configurable_selectors:
            # Set self.config_attrs and delete ID from it
            utils.set_config_attrs(self, data, "config_attrs", self.indices)

        # Set original attributes
        self._original_attributes = data
        # Remove ID
        if "asn" in self._original_attributes:
            self._original_attributes.pop("asn")
        # Remove ID
        if "bgp_neighbors" in self._original_attributes:
            self._original_attributes.pop("bgp_neighbors")
        # Remove ID
        if "aggregate_addresses" in self._original_attributes:
            self._original_attributes.pop("aggregate_addresses")

        # Sets object as materialized
        # Information is loaded from the Device
        self.materialized = True

        # Get BGP Neighbors
        if self.bgp_neighbors == []:
            # Set BGP Neighbor if any
            # Adds BGP Neighbor to parent BGP Router already
            BgpNeighbor.get_all(self.session, self)

        if self.aggregate_addresses == []:
            # Set Aggregate address if any
            # Adds Aggregate Addresses to parent BGP Router already
            AggregateAddress.get_all(self.session, self)
        return True

    @classmethod
    def get_all(cls, session, parent_vrf):
        """
        Perform a GET call to retrieve all system BGP Routers inside a VRF, and
            create a dictionary containing them.

        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param parent_vrf: Vrf object where BGP Router is stored.
        :return: Dictionary containing BGP Router IDs as keys and a BGP Router
            objects as values.
        """
        logging.info("Retrieving all %s data from switch", cls.__name__)

        uri = "{0}/{1}/bgp_routers".format(
            parent_vrf.base_uri, parent_vrf.name
        )

        try:
            response = session.request("GET", uri)
        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        data = json.loads(response.text)

        bgp_dict = {}

        for uri in data.values():
            # Create a BgpRouter object and adds it to parent Vrf object list
            asn, bgp = BgpRouter.from_uri(session, parent_vrf, uri)
            # Load all BGP Router data from within the Switch
            bgp_dict[asn] = bgp

        return bgp_dict

    @PyaoscxModule.connected
    def apply(self):
        """
        Main method used to either create or update an existing BGP Router.
            Checks whether the BGP Router exists in the switch. Calls
            self.update() if BGP Router is being updated. Calls self.create()
            if a new BGP Router is being created.

        :return: Boolean, True if object was created or modified.
        """
        if not self.__parent_vrf.materialized:
            self.__parent_vrf.apply()
        if self.materialized:
            return self.update()
        return self.create()

    @PyaoscxModule.connected
    def update(self):
        """
        Perform a PUT call to apply changes to an existing BGP Router.

        :return: True if Object was modified and a PUT request
            was made.
        """
        bgp_router_data = utils.get_attrs(self, self.config_attrs)
        self.__modified = self._put_data(bgp_router_data)
        return self.__modified

    @PyaoscxModule.connected
    def create(self):
        """
        Perform a POST call to create a new BGP Router table entry. Only
            returns if no exception is raised.

        :return: Boolean, True if entry was created.
        """
        bgp_data = utils.get_attrs(self, self.config_attrs)
        bgp_data["asn"] = self.asn

        self.__modified = self._post_data(bgp_data)
        return self.__modified

    @PyaoscxModule.connected
    def delete(self):
        """
        Perform DELETE call to delete BGP Router.
        """
        self._send_data(self.path, None, "DELETE", "Delete")
        self.__parent_vrf.remove_bgp_router(self)
        # Delete object attributes
        utils.delete_attrs(self, self.config_attrs)

    @classmethod
    def from_response(cls, session, parent_vrf, response_data):
        """
        Create a BgpRouter object given a response_data related to it.

        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param parent_vrf: Vrf object where BGP Router is stored.
        :param response_data: The response must be a dictionary of the form:
            {
            <asn>: "/rest/v10.04/system/vrfs/<name>/bgp_routers/<asn>"
            }
        :return: BgpRouter object.
        """
        bgp_arr = session.api.get_keys(
            response_data, BgpRouter.resource_uri_name
        )
        asn = bgp_arr[0]
        return BgpRouter(session, asn, parent_vrf)

    @classmethod
    def from_uri(cls, session, parent_vrf, uri):
        """
        Create a BgpRouter object given a URI.

        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param parent_vrf: Vrf object where BGP Router is stored.
        :param uri: a String with a URI.
        :return: tuple containing both the BgpRouter object
            and the BGP's asn.
        """
        # Obtain ID from URI
        # system/vrfs/<name>/bgp_routers/<asn>
        asn = uri.split("/")[-1]

        # Create BGP object
        bgp_obj = BgpRouter(session, asn, parent_vrf)

        return asn, bgp_obj

    def __str__(self):
        return "BGP Router ID {0}".format(self.asn)

    @PyaoscxModule.deprecated
    def get_uri(self):
        """
        Method used to obtain the specific BGP Router URI.

        :return: Object's URI.
        """
        return self.path

    @PyaoscxModule.deprecated
    def get_info_format(self):
        """
        Method used to obtain correct object format for referencing inside
            other objects.

        :return: Object format depending on the API Version.
        """
        return self.session.api.get_index(self)

    @property
    def modified(self):
        """
        Return boolean with whether this object has been modified.
        """
        return self.__modified

    @PyaoscxModule.deprecated
    def was_modified(self):
        """
        Getter method for the __modified attribute.

        :return: Boolean True if the object was recently modified.
        """
        return self.modified

    ####################################################################
    # IMPERATIVE FUNCTIONS
    ####################################################################

    def update_bgp_neighbors(self, new_neighbor):
        """
        Update references to BGP Neighbors. If a Neighbor with the same IP,
            Interface or group name is found, update the reference to the new
            neighbor, otherwise, add reference to the new neighbor.

        :param new_neighbor: Neighbor to add reference to.
        """
        for neighbor in self.neighbors:
            if (
                neighbor.ip_or_ifname_or_group_name
                == new_neighbor.ip_or_ifname_or_group_name
            ):
                neighbor = new_neighbor
                return
        self.new_neighbor.append(new_neighbor)

    def remove_bgp_neighbor(self, neighbor):
        """
        Update references to BGP Neighbors. If a Neighbor with the same IP,
            Interface or group name is found, delete the reference to it.

        :param new_neighbor: Neighbor to add reference to.
        """
        for i in self.neighbors:
            if (
                i.ip_or_ifname_or_group_name
                == neighbor.ip_or_ifname_or_group_name
            ):
                self.neighbor.remove(i)

    @PyaoscxModule.materialized
    def create_bgp_neighbors(
        self,
        ip_or_ifname_or_group_name,
        family_type=None,
        activate=None,
        inbound_soft_reconfiguration=None,
        route_reflector_client=None,
        send_community=None,
        remote_as=None,
        update_source=None,
    ):
        """
        Perform a POST call to create BGP Neighbors to the associated current
            BGP Router. With l2vpn_evpn being True, this will also apply EVPN
            settings to the BGP neighbor configurations.

        :param family_type: Alphanumeric to specify what type of neighbor
            settings to configure. The options are 'ipv4-unicast',
            'ipv6-unicast' and 'l2vpn-evpn'.
        :param ip_or_ifname_or_group_name: IPv4 address or name of group of
            the neighbors that functions as the BGP Router link. Example IPv4:
            10.10.12.11/255.255.255.255
        :param activate: Boolean value to activate neighbor in the address
            family.
        :param inbound_soft_reconfiguration: Boolean value that allow inbound
            soft reconfiguration.
        :param route_reflector_client: Boolean value to determine whether this
            neighbor has route reflector enabled.
        :param send_community: Boolean value to determine whether this neighbor
            has send-community enabled.
        :param remote_as: Integer peer ASN.
        :param update_source: Source address for the neighbor session.
        :return: BgpRouter object.
        """
        if (
            activate is not None
            or inbound_soft_reconfiguration is not None
            or route_reflector_client is not None
            or send_community is not None
        ):
            if family_type is None:
                raise VerificationError(
                    "family_type must be provided when at least one of "
                    '"inbound_soft_reconfiguration", "send_community",'
                    '"activate", "route_reflector_client", is enabled'
                )

        _activate = {
            "ipv4-unicast": False,
            "ipv6-unicast": False,
            "l2vpn-evpn": False,
        }

        _inbound_soft_reconfiguration = {
            "ipv4-unicast": False,
            "ipv6-unicast": False,
        }

        _route_reflector_client = {
            "ipv4-unicast": False,
            "ipv6-unicast": False,
            "l2vpn-evpn": False,
        }

        _send_community = {
            "ipv4-unicast": "none",
            "ipv6-unicast": "none",
            "l2vpn-evpn": "none",
        }

        kw = dict()

        if activate is not None:
            _activate[family_type] = True
            kw["activate"] = activate

        if inbound_soft_reconfiguration is not None:
            _inbound_soft_reconfiguration[family_type] = True
            kw["inbound_soft_reconfiguration"] = inbound_soft_reconfiguration

        if route_reflector_client is not None:
            _route_reflector_client[family_type] = True
            kw["route_reflector_client"] = route_reflector_client

        if send_community is not None:
            _send_community[family_type] = send_community
            kw["send_community"] = send_community

        if update_source:
            kw["update_source"] = update_source

        if remote_as:
            kw["remote_as"] = remote_as

        bgp_neighbor_obj = self.session.api.get_module(
            self.session,
            "BgpNeighbor",
            ip_or_ifname_or_group_name,
            parent_bgp_router=self,
            **kw
        )

        # Try to obtain data; if not, create
        try:
            bgp_neighbor_obj.get()
            for key, val in kw.items():
                setattr(bgp_neighbor_obj, key, val)
        except GenericOperationError:
            pass  # not an error, just doesn't exist yet
        finally:
            bgp_neighbor_obj.apply()

        return bgp_neighbor_obj



# (C) Copyright 2023 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

import json
import logging
import re

from pyaoscx.exceptions.generic_op_error import GenericOperationError
from pyaoscx.exceptions.parameter_error import ParameterError
from pyaoscx.exceptions.response_error import ResponseError

import pyaoscx.utils.util as utils

from pyaoscx.pyaoscx_module import PyaoscxModule


class ObjectGroup(PyaoscxModule):
    """
    Provide configuration management for IPv6 on AOS-CX devices.
    """

    indices = ["name", "object_type"]
    collection_uri = "system/acl_object_groups"
    object_uri = collection_uri + "/{name},{type}"
    resource_uri_name = "acl_object_groups"

    def __init__(self, session, name, object_type):
        """
        Create an instance of ObjectGroup class

        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param name: Name of the object group
        :param object_type: Type of object group (ipv4, ipv6, l4port)
        """
        self.session = session
        self.name = name
        self.object_type = object_type
        self.cfg_addresses = {}
        self.cfg_ports = {}
        self.vsx_sync = []
        self.cfg_version = 0
        uri_indices = {
            "name": self.name,
            "type": self.object_type,
        }
        self.base_uri = self.collection_uri
        self.path = self.object_uri.format(**uri_indices)
        self._prev_attrs = {}
        for attr in ["cfg_ports", "cfg_addresses", "vsx_sync"]:
            self._prev_attrs[attr] = getattr(self, attr).copy()

    def __eq__(self, other):
        return (
            isinstance(other, ObjectGroup)
            and self.session == other.session
            and self.name == other.name
            and self.object_type == other.object_type
        )

    @PyaoscxModule.connected
    def get(self, depth=None, selector=None):
        """
        Perform a GET call to retrieve data for an Object Group table entry and
            fill the object with the incoming attributes.

        :param depth: Integer deciding how many levels into the API JSON that
            references will be returned.
        :param selector: Alphanumeric option to select specific information to
            return.
        :return: Returns True if no exception is raised.
        """
        logging.info("Retrieving %s from switch", self)

        depth = depth or self.session.api.default_depth
        selector = selector or self.session.api.default_selector

        self._get_and_copy_data(depth, selector, self.indices)
        for attr in ["cfg_ports", "cfg_addresses", "vsx_sync"]:
            self._prev_attrs[attr] = getattr(self, attr).copy()
        self.materialized = True

    @classmethod
    def get_all(cls, session):
        """
        Perform a GET call to retrieve all Object Groups and create dictionary
            containing each respective class

        :param cls: Object's class
        :param session: pyaoscx.Session object used to represent a logical
            connection to the switch
        :return: Dictionary containing class id and a object as value
        """
        logging.info("Retrieving all %s data from switch", cls.__name__)

        try:
            response = session.request("GET", cls.collection_uri)
        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        data = json.loads(response.text)

        group_collection = {}
        for uri in data.values():
            group_id, group_obj = cls.from_uri(session, uri)
            group_collection[group_id] = group_obj

        return group_collection

    @PyaoscxModule.connected
    def apply(self):
        """
        Main method used to either create or update an
            Object Group table entry

        :return: True if the object was modified
        """
        if self.materialized:
            return self.update()
        return self.create()

    @PyaoscxModule.connected
    def create(self):
        """
        Perform a POST call to create a new Object Group using the object's
            attributes as POST body. Exception is raised if object is unable
            to be created

        :returnd: Boolean, True if entry was created
        """
        logging.info("Creating %s in switch", self)
        group_data = {}
        group_data["cfg_version"] = 0
        group_data["object_type"] = self.object_type
        group_data["name"] = self.name
        self.__modified = self._post_data(group_data)
        return self.__modified

    @PyaoscxModule.connected
    def update(self):
        """
        Perform a PUT call to apply changes to an existing Classifier

        :return: True if object was modified and a PUT request was made.
        """
        group_data = {}
        modified = False
        if (
            self.object_type in ["ipv4", "ipv6"]
            and self._prev_attrs["cfg_addresses"] != self.cfg_addresses
        ):
            group_data["cfg_addresses"] = self.cfg_addresses.copy()
            self.cfg_version = self.cfg_version + 1
            modified = True
        elif (
            self.object_type == "l4port"
            and self._prev_attrs["cfg_ports"] != self.cfg_ports
        ):
            group_data["cfg_ports"] = self.cfg_ports.copy()
            self.cfg_version = self.cfg_version + 1
            modified = True
        if self.vsx_sync != self._prev_attrs["vsx_sync"]:
            group_data["vsx_sync"] = self.vsx_sync
            modified = True
        if modified:
            logging.info("Updating %s in switch", self)
            group_data["cfg_version"] = self.cfg_version
            self.__modified = self._put_data(group_data)
            del group_data["cfg_version"]
            self._prev_attrs.update(group_data)
        else:
            self.__modified = False
        return self.__modified

    @PyaoscxModule.connected
    def delete(self):
        """
        Perform a DELETE call to erase the Classifier entry.
        """
        logging.info("Deleting %s from switch", self)
        self._send_data(self.path, None, "DELETE", "Delete")
        self._prev_attrs.clear()

    @classmethod
    def from_response(cls, session, response_data):
        """
        Create an Onbject Group given a response_data

        :param cls: Object's class
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device
        :param response_data: The response must be a dictionary of the form:
            {id: URL}
        :return: Object Group
        """
        uri = next(iter(response_data.values()))
        _, group_obj = cls.from_uri(session, uri)
        return group_obj

    @classmethod
    def from_uri(cls, session, uri):
        """
        Create an object given an Object Group URI

        :param cls: Object's class
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device
        :param uri: A String with a URI
        :return: tuple containing the Object Group id and the object.
        """
        index_pattern = re.compile(
            r"(.*)system/acl_object_groups/(?P<name>.+),(?P<grp_type>.+)"
        )
        pattern_match = index_pattern.match(uri)
        group_name = pattern_match.group("name")
        group_type = pattern_match.group("grp_type")
        group_key = "{0},{1}".format(group_name, group_type)
        group_obj = cls(session, group_name, group_type)
        return group_key, group_obj

    def __str__(self):
        return "Object Group {0},{1}".format(self.name, self.object_type)

    @PyaoscxModule.deprecated
    def get_uri(self):
        """
        Method used to obtain the specific Object Group URI

        :return: Object's URI
        """
        uri = "{0}{1}".format(self.session.resource_prefix, self.path)
        return uri

    @PyaoscxModule.deprecated
    def get_info_format(self):
        """
        Method to obtain correct object format for referencing inside
            other objects

        return: Object format depending on the API Version
        """
        return self.session.api.get_index(self)

    @property
    def modified(self):
        """
        Return boolean with whether this object has been modified.
        """
        return self.__modified

    @PyaoscxModule.deprecated
    def was_modified(self):
        """
        Getter method for the __modified attribute.

        :return: Boolean True if the object was recently modified.
        """
        return self.modified

    ####################################################################
    # IMPERATIVE FUNCTIONS
    ####################################################################

    @PyaoscxModule.materialized
    def update_ip_entry_to_group(self, index, ip_address):
        """
        Create or update an ip address entry

        :param index: Index of the ip address in the group
        :param ip_address: IP address to add or update
        :return: True if the object has been modified
        """
        version = utils.get_ip_version(ip_address)
        if version != self.object_type:
            raise ParameterError(
                "IP version mismatch: {0} is not {1}".format(
                    ip_address, self.object_type
                )
            )
        if not isinstance(index, str):
            index = str(index)
        group_data = {}
        group_data[index] = utils.fix_ip_mask(ip_address, self.object_type)
        self.cfg_addresses.update(group_data)
        return self.update()

    @PyaoscxModule.materialized
    def remove_ip_entry_from_group(self, index):
        """
        Remove an IP entry from group

        :param index: Index of the ip address to remove
        :return: True if the object has been modified
        """
        if not isinstance(index, str):
            index = str(index)
        if index in self.cfg_addresses:
            del self.cfg_addresses[index]
            return self.update()
        else:
            logging.warning(
                "Index %s not found in IP group for %s", str(index), self
            )
            return False

    @PyaoscxModule.materialized
    def update_port_range_to_group(self, index, port_min=0, port_max=65535):
        """
        Create or update an L4 port range entry

        :param index: Index of the port range in the group
        :param port_min: Minimum port of the range
        :param port_max: Maximum port of the range
        :return: True if the object has been modified
        """
        if self.object_type != "l4port":
            raise ParameterError("{0} is not type L4 port".format(self))
        if not isinstance(index, str):
            index = str(index)
        group_data = {}
        group_data[index] = {
            "l4_port_min": int(port_min),
            "l4_port_max": int(port_max),
        }
        self.cfg_ports.update(group_data)
        return self.update()

    @PyaoscxModule.materialized
    def update_port_to_group(self, index, port):
        """
        Create or update an L4 port entry, it can be numeric or string value

        :param index: Index of the port entry in the group
        :param port: L4 port, it can be string or numeric
        :return: True if the object has been modified
        """
        if not isinstance(index, str):
            index = str(index)
        if isinstance(port, int):
            tmp_port = port
        elif isinstance(port, str):
            if port.isnumeric():
                tmp_port = int(port)
            elif port in utils.l4_ports:
                tmp_port = utils.l4_ports[port]
            else:
                raise ParameterError(
                    "Invalid L4 port name {0}, valid names are: {1}".format(
                        port, ", ".join(utils.l4_ports)
                    )
                )
        else:
            raise ParameterError("Invalid port {0}".format(port))
        group_data = {}
        group_data[index] = {"l4_port_min": tmp_port, "l4_port_max": tmp_port}
        self.cfg_ports.update(group_data)
        return self.update()

    @PyaoscxModule.materialized
    def remove_port_range_from_group(self, index):
        """
        Remove a Port Range entry

        :param index: Index of the port range to remove
        :return: True if the object has been modified
        """
        if not isinstance(index, str):
            index = str(index)
        if index in self.cfg_ports:
            del self.cfg_ports[index]
            return self.update()
        else:
            logging.warning(
                "Index %s not found in Port Range for %s", str(index), self
            )
            return False



# (C) Copyright 2023 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

import json
import logging
import re

from urllib.parse import quote_plus, unquote_plus

from pyaoscx.exceptions.generic_op_error import GenericOperationError
from pyaoscx.exceptions.response_error import ResponseError

from pyaoscx.utils import util as utils

from pyaoscx.pyaoscx_module import PyaoscxModule


class LLDPNeighbor(PyaoscxModule):
    """
    Provide information from LLDP_Neighbor on AOS-CX devices.
    """

    collection_uri = "system/interfaces/{interface}/lldp_neighbors"
    object_uri = collection_uri + "/{chassis_id},{port_id}"
    indices = ["chassis_id", "port_id"]
    resource_uri_name = "lldp_neighbors"

    def __init__(self, session, chassis_id, port_id, parent_interface):
        self.session = session
        self.chassis_id = chassis_id
        self.port_id = port_id
        self._parent_interface = (
            session.api.get_module(session, "Interface", parent_interface)
            if isinstance(parent_interface, str)
            else parent_interface
        )
        self.materialized = False
        self.__modified = False
        uri_indices = {
            "interface": self._parent_interface.percents_name,
            "chassis_id": quote_plus(chassis_id),
            "port_id": quote_plus(port_id),
        }
        self.collection_uri = self.collection_uri.format(**uri_indices)
        self.object_uri = self.object_uri.format(**uri_indices)
        self.base_uri = self.collection_uri
        self.path = self.object_uri

    @PyaoscxModule.connected
    def get(self, depth=None, selector="status"):
        """
        Perform a GET call to retrieve data for a LLDP Neighbor table entry and
            fill the object with the incoming attributes.

        :param depth: Integer deciding how many levels into the API JSON that
            references will be returned.
        :param selector: Alphanumeric option to select specific information to
            return.
        :return: Returns True if no exception is raised.
        """
        logging.info("Retrieving %s from switch", self)

        depth = depth or self.session.api.default_depth
        selector = selector or self.session.api.default_selector

        self._get_and_copy_data(depth, selector, self.indices)
        self.materialized = True
        return True

    @classmethod
    def get_all(cls, session, parent_interface):
        """
        Perform a GET call to retrieve all LLDP Neighbors for an interface
            and create a dictionary containing each LLDP Neighbor.

        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param parent_interface: parent Interface object where LLDP Neighbor
            is stored.
        :return: Dictionary containing LLDP Neighbor's name as key and a LLDP
            Neighbor objects as values.
        """
        logging.info("Retrieving all %s data from switch", cls.__name__)
        if isinstance(parent_interface, str):
            parent_interface = session.api.get_module(
                session, "Interface", parent_interface
            )
        try:
            uri = cls.collection_uri.format(
                interface=parent_interface.percents_name
            )
            response = session.request("GET", uri)
        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)
        data = json.loads(response.text)

        lldp_n_dict = {}
        # Get all URI elements in the form of a list
        uri_list = session.api.get_uri_from_data(data)

        for uri in uri_list:
            # Create an LLDP Neighbor object
            name, lldp_n_obj = cls.from_uri(session, uri)
            lldp_n_dict[name] = lldp_n_obj

        return lldp_n_dict

    @classmethod
    def from_response(cls, session, response_data):
        """
        Create an LLDP Neighbor object given a response_data related to the
            LLDP Neighbor object.

        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param response_data: The response must be a dictionary of the form:
            { "chassis_id,port_id": URL }, with URL:
            ".../system/interfaces/<interface_name>/lldp_neighbors/<c,p>"
        :return: LLDPNeighbor object.
        """
        lldp_n_uri = next(iter(response_data))
        _, lldp_n_obj = cls.from_uri(session, lldp_n_uri)
        return lldp_n_obj

    @classmethod
    def from_uri(cls, session, uri):
        """
        Create an LLDP Neighbor object given a URI.

        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param uri: a String with a URI.
        :return: tuple containing both the LLDP Neighbor's name and LLDP
            Neighbor object object.
        """
        # Obtain ID from URI
        index_pattern = re.compile(
            r"(.*)/(?P<intf>.+)/lldp_neighbors/(?P<chassis>.+),(?P<port_id>.+)"
        )
        intf_name = index_pattern.match(uri).group("intf")
        chassis_id = index_pattern.match(uri).group("chassis")
        port_id = index_pattern.match(uri).group("port_id")
        # Create LLDP Neighbor object
        lldp_neighbor_obj = cls(
            session,
            unquote_plus(chassis_id),
            unquote_plus(port_id),
            unquote_plus(intf_name),
        )

        return "{0},{1}".format(chassis_id, port_id), lldp_neighbor_obj

    @classmethod
    def get_facts(cls, session):
        """
        Perform a GET call to retrieve all LLDP Neighbors and their data.

        :param cls: Class reference.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :return facts: Dictionary containing Interface IDs as keys and
            LLDP Neighbors objects as values.
        """
        logging.info("Retrieving the switch LLDP Neighbors facts")

        # Set depth
        depth = session.api.default_facts_depth

        # Build URI to get all LLDP Neighbors from all Interfaces
        uri = cls.collection_uri.format(interface=quote_plus("*"))
        uri += "?depth={0}".format(depth)

        try:
            # Try to get facts via GET method
            response = session.request("GET", uri)

        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        # Load into json format
        facts = json.loads(response.text)
        return facts

    @PyaoscxModule.connected
    def apply(self):
        """
        Not applicable for LLDP Neighbor.
        """
        pass

    @PyaoscxModule.connected
    def update(self):
        """
        Not applicable for LLDP Neighbor.
        """
        pass

    @PyaoscxModule.connected
    def create(self):
        """
        Not applicable for LLDP Neighbor.
        """
        pass

    @PyaoscxModule.connected
    def delete(self):
        """
        Not applicable for LLDP Neighbor.
        """
        pass

    def __str__(self):
        return "LLDP Neighbor for interface {0}: {1},{2}".format(
            self._parent_interface.name, self.chassis_id, self.port_id
        )

    @PyaoscxModule.deprecated
    def get_info_format(self):
        """
        Method used to obtain correct object format for referencing inside
            other objects.

        return: Object format depending on the API Version.
        """
        return self.session.api.get_index(self)



# (C) Copyright 2019-2024 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

import json
import logging
import re

from pyaoscx.exceptions.generic_op_error import GenericOperationError
from pyaoscx.exceptions.parameter_error import ParameterError
from pyaoscx.exceptions.response_error import ResponseError
from pyaoscx.exceptions.verification_error import VerificationError

from pyaoscx.utils import util as utils

from pyaoscx.pyaoscx_module import PyaoscxModule


class AclEntry(PyaoscxModule):
    """
    Provide configuration management for ACL Entry on AOS-CX devices.
    """

    indices = ["sequence_number"]
    resource_uri_name = "cfg_aces"

    # Lists of fields according SW capablities
    cap_dscp = ["dscp"]
    cap_ecn = ["ecn"]
    cap_frg = ["fragment"]
    cap_tcp_flags = [
        "tcp_ack",
        "tcp_cwr",
        "tcp_ece",
        "tcp_established",
        "tcp_fin",
        "tcp_psh",
        "tcp_rst",
        "tcp_syn",
        "tcp_urg",
    ]
    cap_mac = [
        "dst_mac",
        "src_mac",
    ]
    cap_grp = [
        "src_ip_group",
        "dst_ip_group",
        "src_l4_port_group",
        "dst_l4_port_group",
    ]
    cap_pre = ["ip_precedence"]
    cap_pcp = ["pcp"]
    cap_tos = ["tos"]
    cap_ttl = ["ttl"]
    cap_log = ["log"]
    cap_l4_port = [
        "dst_l4_port_max",
        "dst_l4_port_min",
        "src_l4_port_max",
        "src_l4_port_min",
    ]

    # These parameters cannot be changed once the ACE is created
    # If any of them is set for update the entire ACE must be
    # replaced (delete and re-create). This list is needed while
    # we are not able to read the schema.
    immutable_parameter_names = (
        [
            "action",
            "count",
            "dst_ip",
            "ethertype",
            "icmp_code",
            "icmp_type",
            "log",
            "protocol",
            "sequence_number",
            "src_ip",
            "vlan",
        ]
        + cap_l4_port
        + cap_dscp
        + cap_ecn
        + cap_frg
        + cap_tcp_flags
        + cap_mac
        + cap_grp
        + cap_pre
        + cap_pcp
        + cap_tos
        + cap_ttl
        + cap_log
    )

    # This list needs to be maintained only while we are not
    # able to read the schema. It is required when extracting
    # parameters to create copies.
    mutable_parameter_names = ["comment"]
    setter_attr = [
        "src_ip",
        "dst_ip",
        "protocol",
        "src_mac",
        "dst_mac",
        "dscp",
        "ethertype",
        "icmp_type",
        "src_l4_port_min",
        "src_l4_port_max",
        "dst_l4_port_min",
        "dst_l4_port_max",
    ]

    def __init__(
        self, session, sequence_number, parent_acl, uri=None, **kwargs
    ):

        self.session = session
        # Assign ID
        self.sequence_number = sequence_number
        # Assign parent Acl object
        self.__set_acl(parent_acl)
        self._uri = uri
        # List used to determine attributes related to the acl_entry
        # configuration
        self.config_attrs = []
        self.materialized = False
        # Attribute dictionary used to manage the original data
        # obtained from the GET
        self.__original_attributes = {}
        eval_args = {}
        for new_attr in self.setter_attr:
            if new_attr in kwargs:
                setattr(self, new_attr, kwargs[new_attr])
                eval_args[new_attr] = kwargs.pop(new_attr)

        # Checking against C&C
        _exclude_args = []
        _not_supported = []
        if parent_acl.list_type == "ipv6":
            if "classifier_ace_v6_ttl" not in parent_acl.capabilities:
                _exclude_args.append("ttl")
            if "classifier_ace_v6_vlan_id" not in parent_acl.capabilities:
                _exclude_args.append("vlan")
            if "classifier_ace_v6_tcp_flg" not in parent_acl.capabilities:
                _exclude_args.extend(self.cap_tcp_flags)
        if "classifier_ace_dscp" not in parent_acl.capabilities:
            _exclude_args.extend(self.cap_dscp)
        if "classifier_ace_ecn" not in parent_acl.capabilities:
            _exclude_args.extend(self.cap_ecn)
        if "classifier_ace_frg" not in parent_acl.capabilities:
            _exclude_args.extend(self.cap_frg)
        if "classifier_ace_tcp_flags" not in parent_acl.capabilities:
            _exclude_args.extend(self.cap_tcp_flags)
        if "classifier_ace_tcp_flg_cwr" not in parent_acl.capabilities:
            _exclude_args.append("tcp_cwr")
        if "classifier_ace_tcp_flg_ece" not in parent_acl.capabilities:
            _exclude_args.append("tcp_ece")
        if "classifier_ace_pre" not in parent_acl.capabilities:
            _exclude_args.extend(self.cap_pre)
        if "classifier_acl_object_group" not in parent_acl.capabilities:
            _exclude_args.extend(self.cap_grp)
        if "classifier_ace_pcp" not in parent_acl.capabilities:
            _exclude_args.extend(self.cap_pcp)
        if "classifier_ace_tos" not in parent_acl.capabilities:
            _exclude_args.extend(self.cap_tos)
        if "classifier_ace_ttl" not in parent_acl.capabilities:
            _exclude_args.extend(self.cap_ttl)
        if "action" not in kwargs or kwargs["action"] == "permit":
            if "classifier_acl_log_permit" not in parent_acl.capabilities:
                _exclude_args.extend(self.cap_log)
        elif kwargs["action"] == "deny":
            if "classifier_acl_log_deny" not in parent_acl.capabilities:
                _exclude_args.extend(self.cap_log)

        if (
            parent_acl.list_type == "ipv4"
            and "protocol" in eval_args
            and eval_args["protocol"] in ["ah", "51", 51]
            and "classifier_ace_v4_ah_ingress" not in parent_acl.capabilities
            and "classifier_ace_v4_ah_egress" not in parent_acl.capabilities
        ):
            _not_supported.append("protocol='ah'")
        if (
            "protocol" in eval_args
            and eval_args["protocol"] in ["sctp", "132", 132]
            and [
                l4_parm for l4_parm in self.cap_l4_port if l4_parm in eval_args
            ]
            != []
            and "classifier_ace_sctp_l4_port" not in parent_acl.capabilities
        ):
            _not_supported.append("protocol='sctp' with L4 src/dst")
        for arg in _exclude_args:
            if arg in kwargs:
                _not_supported.append(arg)
        if _not_supported != []:
            raise ParameterError(
                "[ACL {0}/{1} - Entry {2}] Parameters not supported: {3}".format(
                    parent_acl.name,
                    parent_acl.list_type,
                    sequence_number,
                    ", ".join(_not_supported),
                )
            )
        # Set arguments needed for correct creation
        utils.set_creation_attrs(self, **kwargs)
        # Attribute used to know if object was changed recently
        self.__modified = False

    def __eq__(self, other):
        return (
            isinstance(other, AclEntry)
            and self.session == other.session
            and self.sequence_number == other.sequence_number
            and self.base_uri == other.base_uri
        )

    def __set_acl(self, parent_acl):
        """
        Set parent Acl object as an attribute for the AclEntry object.

        :param parent_acl: a Pyaoscx.Acl object.
        """
        # Set parent acl
        self.__parent_acl = parent_acl

        # Set URI
        self.base_uri = "{0}/{1}{2}{3}/cfg_aces".format(
            self.__parent_acl.base_uri,
            self.__parent_acl.name,
            self.session.api.compound_index_separator,
            self.__parent_acl.list_type,
        )

        if self.sequence_number not in self.__parent_acl.cfg_aces:
            self.__parent_acl.cfg_aces[self.sequence_number] = self

    @PyaoscxModule.connected
    def get(self, depth=None, selector="configuration"):
        """
        Perform a GET call to retrieve data for an ACL Entry table entry and
            fill the object with the incoming attributes.

        :param depth: Integer deciding how many levels into the API JSON that
            references will be returned.
        :param selector: Alphanumeric option to select specific information to
            return.
        :return: Returns True if no exception is raised.
        """
        logging.info("Retrieving %s from switch", self)

        depth = depth or self.session.api.default_depth
        selector = selector or self.session.api.default_selector

        if not self.session.api.valid_depth(depth):
            depths = self.session.api.valid_depths
            raise Exception("ERROR: Depth should be {0}".format(depths))

        if selector not in self.session.api.valid_selectors:
            selectors = " ".join(self.session.api.valid_selectors)
            raise Exception(
                "ERROR: Selector should be one of {0}".format(selectors)
            )

        payload = {"depth": depth, "selector": selector}

        uri = "{0}/{1}".format(self.base_uri, self.sequence_number)
        try:
            response = self.session.request("GET", uri, params=payload)

        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        get_data = json.loads(response.text)
        orig_data = {k: v for k, v in get_data.items() if v is not None}
        ObjectGroup = self.session.api.get_module_class(
            self.session, "ObjectGroup"
        )
        for grp_attr in self.cap_grp:
            if grp_attr in orig_data:
                obj_grp = ObjectGroup.from_response(
                    self.session, orig_data[grp_attr]
                )
                orig_data[grp_attr] = obj_grp
        data = orig_data.copy()
        for new_attr in [
            "src_ip",
            "dst_ip",
            "protocol",
            "src_mac",
            "dst_mac",
            "dscp",
            "ethertype",
            "icmp_type",
            "src_l4_port_min",
            "src_l4_port_max",
            "dst_l4_port_min",
            "dst_l4_port_max",
        ]:
            if new_attr in data:
                setattr(self, new_attr, data.pop(new_attr))
        # Add dictionary as attributes for the object
        utils.create_attrs(self, data)

        # Determines if the ACL Entry is configurable
        if selector in self.session.api.configurable_selectors:
            # Set self.config_attrs and delete ID from it
            utils.set_config_attrs(
                self, data, "config_attrs", ["sequence_number"]
            )

        # Set original attributes
        self.__original_attributes.update(orig_data)
        # Remove ID
        if "sequence_number" in self.__original_attributes:
            self.__original_attributes.pop("sequence_number")

        # Sets object as materialized
        # Information is loaded from the Device
        self.materialized = True
        return True

    @classmethod
    def get_all(cls, session, parent_acl):
        """
        Perform a GET call to retrieve all system ACL Entries inside an ACL,
            and create a dictionary containing them.

        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param parent_acl: parent Acl object where ACL Entry is stored.
        :return acl_entry_dict: Dictionary containing ACL Entry IDs as keys
            and an ACL Entry objects as values.
        """
        logging.info("Retrieving all %s data from switch", cls.__name__)
        uri = "{0}/{1}{2}{3}/cfg_aces?depth={4}".format(
            parent_acl.base_uri,
            parent_acl.name,
            session.api.compound_index_separator,
            parent_acl.list_type,
            session.api.default_facts_depth,
        )

        try:
            response = session.request("GET", uri)
        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        data = json.loads(response.text)

        acl_entry_dict = {}

        ObjectGroup = session.api.get_module_class(session, "ObjectGroup")
        for ace_seq_num, ace_dict in data.items():
            # Create a AclEntry object and setting attributes from response
            ace_seq_num = int(ace_seq_num)
            ace_kwargs = {k: v for k, v in ace_dict.items() if v is not None}
            if "sequence_number" in ace_kwargs:
                del ace_kwargs["sequence_number"]
            orig_keys = list(ace_kwargs.keys())
            for grp_attr in cls.cap_grp:
                if grp_attr in ace_kwargs:
                    obj_grp = ObjectGroup.from_response(
                        session, ace_kwargs[grp_attr]
                    )
                    ace_kwargs[grp_attr] = obj_grp
            ace_obj = cls(session, ace_seq_num, parent_acl, **ace_kwargs)
            ace_orig = utils.get_attrs(ace_obj, orig_keys)
            ace_obj.__original_attributes.update(ace_orig)
            ace_obj.config_attrs = list(ace_orig.keys())
            ace_obj.materialized = True
            acl_entry_dict[ace_seq_num] = ace_obj

        return acl_entry_dict

    @PyaoscxModule.connected
    def apply(self):
        """
        Main method used to either create a new ACL Entry or update an existing
            one. It is possible that in case the are differences between the
            ACE on the switch and the local representation on immutable
            attributes a replace (delete+create) will take place. Note that
            unspecified parameters will be kept intact.

        :return modified: Boolean, True if object was created or modified.
        """
        if not self.__parent_acl.materialized:
            self.__parent_acl.get()

        modified = False

        remote_ace = AclEntry(
            self.session,
            self.sequence_number,
            self.__parent_acl,
            **self.__original_attributes
        )

        if not self.materialized:
            modified = self.create()
        else:
            if PyaoscxModule._is_replace_required(
                current=remote_ace,
                replacement=self,
                immutable_parameter_names=self.immutable_parameter_names,
            ):
                remote_ace.delete()
                logging.info(
                    "%s for %s will be replaced", self, self.__parent_acl.name
                )
                modified = self.create()
            else:
                # A replace was not required, so it is possible that an
                # update will suffice. Extracting the parameters from
                # the remote works as a materialization
                self.materialized = True
                modified = self.update()

        return modified

    @PyaoscxModule.connected
    def update(self):
        """
        Perform a PUT call to apply changes to an existing ACL Entry.

        :return modified: True if Object was modified and a PUT request
            was made.
        """
        # Variable returned
        modified = False

        acl_entry_data = utils.get_attrs(self, self.config_attrs)

        uri = "{0}/{1}".format(self.base_uri, self.sequence_number)

        if not self.__was_modified():
            # Object was not modified
            modified = False

        else:
            # Normally objects will pull from the switch only writable
            # attributes, but because of the replace (delete/create) behaviour
            # implemented for ACL Entries, we usually pull all configurable
            # attributes. Which means that hey have to be removed from the
            # request before being sent to the switch
            for key in list(acl_entry_data):
                if key in self.immutable_parameter_names:
                    del acl_entry_data[key]

            post_data = json.dumps(acl_entry_data)

            try:
                response = self.session.request("PUT", uri, data=post_data)

            except Exception as e:
                raise ResponseError("PUT", e)

            if not utils._response_ok(response, "PUT"):
                raise GenericOperationError(
                    response.text, response.status_code
                )

            logging.info("SUCCESS: Updating %s", self)
            # Set new original attributes
            self.__original_attributes = acl_entry_data

            # Object was modified
            modified = True

        if modified:
            self.__parent_acl.get()
            self.__parent_acl.apply()

        self.__modified = modified
        return modified

    @PyaoscxModule.connected
    def create(self):
        """
        Perform a POST call to create a new ACL Entry. Only returns if no
            exception is raised.

        :return modified: Boolean, True if entry was created
        """
        acl_entry_data = utils.get_attrs(self, self.config_attrs)
        acl_entry_data["sequence_number"] = self.sequence_number

        for new_attr in [
            "src_ip",
            "dst_ip",
            "src_mac",
            "dst_mac",
            "protocol",
            "ethertype",
        ]:
            new_value = getattr(self, new_attr)
            if new_value == "any":
                acl_entry_data.pop(new_attr, None)
            else:
                acl_entry_data[new_attr] = new_value

        for new_attr in [
            "dscp",
            "icmp_type",
            "src_l4_port_min",
            "src_l4_port_max",
            "dst_l4_port_min",
            "dst_l4_port_max",
        ]:
            new_value = getattr(self, new_attr)
            if new_value:
                acl_entry_data[new_attr] = new_value

        for group_param in self.cap_grp:
            if group_param in acl_entry_data:
                group_obj = acl_entry_data[group_param]
                acl_entry_data[group_param] = group_obj.get_info_format()

        post_data = json.dumps(acl_entry_data)

        try:
            response = self.session.request(
                "POST", self.base_uri, data=post_data
            )

        except Exception as e:
            raise ResponseError("POST", e)

        if not utils._response_ok(response, "POST"):
            raise GenericOperationError(response.text, response.status_code)

        logging.info("SUCCESS: Adding %s", self)

        # Get all object's data
        self.get()
        self.__parent_acl.get()
        self.__parent_acl.apply()

        # Object was created, means modified
        self.__modified = True
        return True

    @PyaoscxModule.connected
    def delete(self):
        """
        Perform DELETE call to delete ACL Entry from parent ACL on the switch.
        """
        uri = "{0}/{1}".format(self.base_uri, self.sequence_number)

        try:
            response = self.session.request("DELETE", uri)

        except Exception as e:
            raise ResponseError("DELETE", e)

        if not utils._response_ok(response, "DELETE"):
            raise GenericOperationError(response.text, response.status_code)

        logging.info("SUCCESS: Deleting %s", self)

        # Delete back reference from ACL
        del self.__parent_acl.cfg_aces[self.sequence_number]

        self.__parent_acl.get()
        self.__parent_acl.apply()

        # Delete object attributes
        for attr in self.setter_attr:
            if attr in self.config_attrs:
                self.config_attrs.remove(attr)

        utils.delete_attrs(self, self.config_attrs)

    @classmethod
    def from_response(cls, session, parent_acl, response_data):
        """
        Create a AclEntry object given a response_data related to the ACL Entry
            sequence_number object.

        :param cls: Class calling the method.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param parent_acl: parent Acl object where ACL Entry is stored.
        :param response_data: The response must be a dictionary of the form:
            {
            <seq_number>: "/rest/v10.04/system/acls/cfg_aces/<seq_number>"
            }
        :return: AclEntry object.
        """
        acl_entry_arr = session.api.get_keys(
            response_data, AclEntry.resource_uri_name
        )
        sequence_number = acl_entry_arr[0]
        return AclEntry(session, sequence_number, parent_acl)

    @classmethod
    def from_uri(cls, session, parent_acl, uri):
        """
        Create a AclEntry object given a URI.

        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param parent_acl: parent Acl object where ACL Entry is stored.
        :param uri: a String with a URI.
        :return index, acl_entry_obj: tuple containing both the AclEntry
            object and the acl_entry's sequence_number.
        """
        # Obtain ID from URI
        index_pattern = re.compile(r"(.*)cfg_aces/(?P<index>.+)")
        index = index_pattern.match(uri).group("index")

        # Create AclEntry object
        acl_entry_obj = AclEntry(session, index, parent_acl, uri=uri)

        return index, acl_entry_obj

    def __str__(self):
        return "ACL Entry ID {0}".format(self.sequence_number)

    @PyaoscxModule.deprecated
    def get_uri(self):
        """
        Method used to obtain the specific ACL Entry URI.

        return: AclEntry object's URI.
        """
        if self._uri is None:
            self._uri = "{0}{1}/{2}".format(
                self.session.resource_prefix,
                self.base_uri,
                self.sequence_number,
            )

        return self._uri

    @PyaoscxModule.deprecated
    def get_info_format(self):
        """
        Method used to obtain correct object format for referencing inside
            other objects.

        return: AclEntry object format depending on the API Version.
        """
        return self.session.api.get_index(self)

    @property
    def modified(self):
        return self.__modified

    def __was_modified(self):
        """
        Determine if the object was modified since the last materialization.
        """
        current = utils.get_attrs(self, self.config_attrs)
        original = self.__original_attributes
        if current == original:
            return False
        # Because ACL gets all configurable parameters and not just the
        # writable ones those two dictionaries could be different but because
        # on is missing members that the other one has. So it is necessary to
        # check that the missing members are set as None in the other.
        modified = False
        for key, value in current.items():
            if key in original:
                modified |= value != original[key]
            else:
                modified |= value is not None
        for key, value in original.items():
            if key in current:
                modified |= value != current[key]
            else:
                modified |= value is not None

        return modified

    @PyaoscxModule.deprecated
    def was_modified(self):
        """
        Getter method for the __modified attribute.

        :return: Boolean True if the object was recently modified.
        """
        return self.modified

    ####################################################################
    # IMPERATIVE FUNCTIONS
    ####################################################################

    def modify(
        self,
        action=None,
        count=None,
        src_ip=None,
        dst_ip=None,
        dst_l4_port_min=None,
        dst_l4_port_max=None,
        src_mac=None,
        dst_mac=None,
        ethertype=None,
    ):
        """
        Create an AclEntry object, ACL Entry already exists, value passed won't
            update the entry.

        :param action: Action should be either "permit" or "deny".
        :param count: Optional boolean flag that when true, will make entry
            increment hit count for matched packets.
        :param src_ip: Optional source IP address. Both IPv4 and IPv6 are
            supported. Example:
            10.10.12.11/255.255.255.255
            2001:db8::11/ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff
        :param dst_ip: Optional destination IP address. Both IPv4 and IPv6 are
            supported. Example:
            10.10.12.11/255.255.255.255
            2001:db8::11/ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff
        :param dst_l4_port_min: Optional minimum L4 port number in range; used
            in conjunction with dst_l4_port_max.
        :param dst_l4_port_max: Optional maximum L4 port number in range; used
            in conjunction with dst_l4_port_min.
        :param src_mac: Optional source MAC address. Example:
            '01:02:03:04:05:06'
        :param dst_mac: Optional destination MAC address. Example:
            '01:02:03:04:05:06'
        :param ethertype: Optional integer EtherType number.
        :return: True if object was changed.
        """
        if action is not None:
            self.action = action

        if count is not None:
            self.count = count

        if src_ip is not None:
            self.src_ip = src_ip

        if dst_ip is not None:
            self.dst_ip = dst_ip

        if dst_l4_port_min is not None:
            self.dst_l4_port_min = dst_l4_port_min

        if dst_l4_port_max is not None:
            self.dst_l4_port_max = dst_l4_port_max

        if src_mac is not None:
            self.src_mac = src_mac

        if dst_mac is not None:
            self.dst_mac = dst_mac

        if ethertype is not None:
            self.ethertype = ethertype

        # Apply changes
        return self.apply()

    @property
    def src_mac(self):
        """
        Getter for src_mac attribute

        :return: Source MAC address
        """
        return self._src_mac if hasattr(self, "_src_mac") else "any"

    @src_mac.setter
    def src_mac(self, new_src_mac):
        """
        Setter for src_mac attribute
        """
        if new_src_mac and new_src_mac.lower() != "any":
            acl_type = self.__parent_acl.list_type
            if acl_type != "mac":
                raise ParameterError(
                    "MAC Address not allowed for ACL type {0}".format(acl_type)
                )
            self._src_mac = utils.validate_mac_address(
                new_src_mac, cisco_format=True
            )
        else:
            self._src_mac = "any"

    @property
    def dst_mac(self):
        """
        Getter for dst_mac attribute

        :return: Destination MAC address
        """
        return self._dst_mac if hasattr(self, "_dst_mac") else "any"

    @dst_mac.setter
    def dst_mac(self, new_dst_mac):
        """
        Setter for dst_mac attribute
        """
        if new_dst_mac and new_dst_mac.lower() != "any":
            acl_type = self.__parent_acl.list_type
            if acl_type != "mac":
                raise ParameterError(
                    "MAC Address not allowed for ACL type {0}".format(acl_type)
                )
            self._dst_mac = utils.validate_mac_address(
                new_dst_mac, cisco_format=True
            )
        else:
            self._dst_mac = "any"

    @property
    def src_ip(self):
        """
        Getter method for source ip attribute.

        :return: String value for src_ip.
        """
        return self._src_ip if hasattr(self, "_src_ip") else "any"

    @src_ip.setter
    def src_ip(self, new_src_ip):
        """
        Setter method for the src_ip attribute.
        """
        if new_src_ip and new_src_ip.lower() != "any":
            version = utils.get_ip_version(new_src_ip)
            if version != self.__parent_acl.list_type:
                raise VerificationError(
                    "Version does not match the IP "
                    "version type in {}".format(self.__parent_acl.name)
                )
            self._src_ip = utils.fix_ip_mask(new_src_ip, version)
        else:
            self._src_ip = "any"

    @property
    def dst_ip(self):
        """
        Getter method for destination ip attribute.

        :return: String value for dst_ip.
        """
        return self._dst_ip if hasattr(self, "_dst_ip") else "any"

    @dst_ip.setter
    def dst_ip(self, new_dst_ip):
        """
        Setter method for the dst_ip attribute.
        """
        if new_dst_ip and new_dst_ip.lower() != "any":
            version = utils.get_ip_version(new_dst_ip)
            if version != self.__parent_acl.list_type:
                raise VerificationError(
                    "Version does not match the IP "
                    "version type in {}".format(self.__parent_acl.name)
                )
            self._dst_ip = utils.fix_ip_mask(new_dst_ip, version)
        else:
            self._dst_ip = "any"

    @property
    def dscp(self):
        """
        Getter method for DSCP attribute.

        :return: DSCP value, integer or String
        """
        return self._dscp if hasattr(self, "_dscp") else None

    @dscp.setter
    def dscp(self, new_dscp):
        """
        Setter method for the dscp attribute.
        """
        if isinstance(new_dscp, str):
            if new_dscp not in utils.dscp:
                raise ParameterError(
                    "Invalid DSCP {0} - valid DSCP values are: {1}".format(
                        new_dscp, ", ".join(utils.dscp)
                    )
                )
            self._dscp = utils.dscp[new_dscp]
        else:
            self._dscp = new_dscp

    @property
    def protocol(self):
        """
        Getter method for protocol attribute

        :return: protocol value, integer or string
        """
        return self._protocol if hasattr(self, "_protocol") else "any"

    @protocol.setter
    def protocol(self, new_proto):
        """
        Setter method for protocol attribute
        """
        if isinstance(new_proto, str):
            if new_proto.lower() in ["ip", "any", "ipv6", ""]:
                self._protocol = "any"
            elif new_proto in utils.ip_protocols:
                self._protocol = utils.ip_protocols[new_proto]
            else:
                raise ParameterError(
                    "Unknown IP protocol {0}, valid protocols: {1}".format(
                        new_proto, ", ".join(utils.ip_protocols)
                    )
                )
        else:
            self._protocol = new_proto

    @property
    def icmp_type(self):
        """
        Getter for icmp_type attribute

        :return: Icmp type value, string or integer
        """
        return self._icmp_type if hasattr(self, "_icmp_type") else None

    @icmp_type.setter
    def icmp_type(self, new_icmp_type):
        acl_type = self.__parent_acl.list_type
        icmp_types_dict = (
            utils.icmp_types if acl_type == "ipv4" else utils.icmpv6_types
        )
        if isinstance(new_icmp_type, str):
            if new_icmp_type in icmp_types_dict:
                self._icmp_type = icmp_types_dict[new_icmp_type]
            else:
                raise ParameterError(
                    "Invalid ICMP Type {0} - valid types are: {1}".format(
                        new_icmp_type, ", ".join(icmp_types_dict)
                    )
                )
        else:
            self._icmp_type = new_icmp_type

    @property
    def ethertype(self):
        """
        Getter for ethertype attribute

        :return: Ethertype value, integer or string
        """
        return self._ethertype if hasattr(self, "_ethertype") else "any"

    @ethertype.setter
    def ethertype(self, new_ethertype):
        """
        Setter for ethertype attribute
        """
        if isinstance(new_ethertype, str):
            if new_ethertype.lower() == "any":
                self._ethertype = "any"
            elif new_ethertype in utils.ethertypes:
                self._ethertype = utils.ethertypes[new_ethertype]
            else:
                raise ParameterError(
                    "Unknown Ethertype {0} - valid ethertypes are: {1}".format(
                        new_ethertype, ", ".join(utils.ethertypes)
                    )
                )
        else:
            self._ethertype = new_ethertype

    @property
    def src_l4_port_min(self):
        """
        Getter for src_l4_port_min attribute

        :return: Source minimum L4 port
        """
        return (
            self._src_l4_port_min
            if hasattr(self, "_src_l4_port_min")
            else None
        )

    @src_l4_port_min.setter
    def src_l4_port_min(self, new_l4_port):
        """
        Setter for src_l4_port_min attribute
        """
        if isinstance(new_l4_port, str):
            if new_l4_port in utils.l4_ports:
                self._src_l4_port_min = utils.l4_ports[new_l4_port]
            else:
                raise ParameterError(
                    "Unknown L4 port {0}, valid ports are: {1}".format(
                        new_l4_port, ", ".join(utils.l4_ports)
                    )
                )
        else:
            self._src_l4_port_min = new_l4_port

    @property
    def src_l4_port_max(self):
        """
        Getter for src_l4_port_max attribute

        :return: Source maximum L4 port
        """
        return (
            self._src_l4_port_max
            if hasattr(self, "_src_l4_port_max")
            else None
        )

    @src_l4_port_max.setter
    def src_l4_port_max(self, new_l4_port):
        """
        Setter for src_l4_port_max attribute
        """
        if isinstance(new_l4_port, str):
            if new_l4_port in utils.l4_ports:
                self._src_l4_port_max = utils.l4_ports[new_l4_port]
            else:
                raise ParameterError(
                    "Unknown L4 port {0}, valid ports are: {1}".format(
                        new_l4_port, ", ".join(utils.l4_ports)
                    )
                )
        else:
            self._src_l4_port_max = new_l4_port

    @property
    def dst_l4_port_min(self):
        """
        Getter for dst_l4_port_min attribute

        :return: Destination minimum L4 port
        """
        return (
            self._dst_l4_port_min
            if hasattr(self, "_dst_l4_port_min")
            else None
        )

    @dst_l4_port_min.setter
    def dst_l4_port_min(self, new_l4_port):
        """
        Setter for dst_l4_port_min attribute
        """
        if isinstance(new_l4_port, str):
            if new_l4_port in utils.l4_ports:
                self._dst_l4_port_min = utils.l4_ports[new_l4_port]
            else:
                raise ParameterError(
                    "Unknown L4 port {0}, valid ports are: {1}".format(
                        new_l4_port, ", ".join(utils.l4_ports)
                    )
                )
        else:
            self._dst_l4_port_min = new_l4_port

    @property
    def dst_l4_port_max(self):
        """
        Getter for dst_l4_port_max attribute

        :return: Destination maximum L4 port
        """
        return (
            self._dst_l4_port_max
            if hasattr(self, "_dst_l4_port_max")
            else None
        )

    @dst_l4_port_max.setter
    def dst_l4_port_max(self, new_l4_port):
        """
        Setter for src_l4_port_max attribute
        """
        if isinstance(new_l4_port, str):
            if new_l4_port in utils.l4_ports:
                self._dst_l4_port_max = utils.l4_ports[new_l4_port]
            else:
                raise ParameterError(
                    "Unknown L4 port {0}, valid ports are: {1}".format(
                        new_l4_port, ", ".join(utils.l4_ports)
                    )
                )
        else:
            self._dst_l4_port_max = new_l4_port



# (C) Copyright 2021-2023 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

import logging

from pyaoscx.exceptions.verification_error import VerificationError
from pyaoscx.pyaoscx_module import PyaoscxModule
from pyaoscx.utils import util as utils


class PoEInterface(PyaoscxModule):
    """
    Provide configuration management for PoE Interface on AOS-CX devices.
    """

    collection_uri = "system/interfaces"
    object_uri = collection_uri + "/{interface}/poe_interface"
    resource_uri_name = "poe_interface"
    indices = ["name"]

    def __init__(self, session, parent_interface, **kwargs):
        """
        Create an instance of PoEInterface Class.

        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param parent_interface: parent Inferface object where PoE is stored.
        """
        self.session = session
        # List used to determine attributes related to the PoEInterface
        # configuration
        self.config_attrs = []
        # Attribute dictionary used to manage the original data
        # obtained from the GET
        self.__original_attributes = {}
        self.materialized = False
        # Set arguments needed for correct creation
        utils.set_creation_attrs(self, **kwargs)
        self.base_uri = self.collection_uri
        if isinstance(parent_interface, str):
            parent_interface = self.session.api.get_module(
                self.session, "Interface", parent_interface
            )
        self.name = parent_interface.name
        self.path = self.object_uri.format(
            interface=parent_interface.percents_name
        )
        self._interface = parent_interface
        # Attribute used to know if object was changed recently
        self.__modified = False

    @PyaoscxModule.connected
    def get(self, depth=None, selector=None):
        """
        Perform a GET call to retrieve data for a PoE Interface table entry and
            fill the object with the incoming attributes.

        :param depth: Integer deciding how many levels into the API JSON that
            references will be returned.
        :param selector: Alphanumeric option to select specific information to
            return.
        :return: Returns True if no exception is raised.
        """
        logging.info("Retrieving %s from switch", self)

        depth = depth or self.session.api.default_depth
        selector = selector or self.session.api.default_selector

        self._get_and_copy_data(depth, selector, self.indices)
        self.materialized = True
        return True

    @PyaoscxModule.connected
    def apply(self):
        """
        Apply an update of values of this PoE Interface. Calls self.update()
            to apply changes to an existing PoE Interface.
        """
        modified = False
        if self.materialized:
            modified = self.update()

        # Set internal attribute
        self.__modified = modified
        return modified

    @PyaoscxModule.connected
    def update(self):
        """
        Perform a PUT call to apply changes to an existing PoE Interface.

        :return: True if Object was modified and a PUT request was
            made.
        """
        poe_interface_data = utils.get_attrs(self, self.config_attrs)
        poe_interface_data["config"] = self.config.copy()
        self.__modified = self._put_data(poe_interface_data)
        logging.info("SUCCESS: Updating %s", self)
        self.__original_attributes = poe_interface_data
        return self.__modified

    @PyaoscxModule.connected
    def create(self):
        """
        Not needed for PoEInterface
        """
        pass

    @PyaoscxModule.connected
    def delete(self):
        """
        Not needed for PoEInterface
        """
        pass

    @PyaoscxModule.connected
    def get_all(self):
        """
        Not needed for PoEInterface
        """
        pass

    @classmethod
    def from_uri(cls):
        """
        Not needed for PoEInterface
        """
        pass

    def __str__(self):
        return "PoE Interface {0}".format(self.name)

    @PyaoscxModule.deprecated
    def get_uri(self):
        """
        Method used to obtain the specific PoE Interface URI.

        :return: Object URI.
        """
        uri = "{0}{1}".format(self.session.resource_prefix, self.path)

        return uri

    @PyaoscxModule.deprecated
    def get_info_format(self):
        pass

    @property
    def modified(self):
        """
        Return boolean with whether this object has been modified.
        """
        return self.__modified

    @PyaoscxModule.deprecated
    def was_modified(self):
        """
        Getter method for the __modified attribute.

        :return: Boolean True if the object was recently modified.
        """
        return self.modified

    @property
    @PyaoscxModule.materialized
    def allocate_by_method(self):
        """
        Getter method for the allocate_by_method attribute.

        :return: String value for method.
        """
        return self.config["allocate_by_method"]

    @allocate_by_method.setter
    @PyaoscxModule.materialized
    def allocate_by_method(self, new_method):
        """
        Setter method for the allocate_by_method attribute.
        """
        valid_methods = ["class", "usage"]
        if new_method not in valid_methods:
            raise VerificationError(
                "Invalid method {0}, method must be one of {1}".format(
                    new_method, ",".join(valid_methods)
                )
            )
        self.config["allocate_by_method"] = new_method

    @property
    @PyaoscxModule.materialized
    def assigned_class(self):
        """
        Getter method for the cfg_assigned_class attribute.

        :return: String value for assigned class.
        """
        return self.config["cfg_assigned_class"]

    @assigned_class.setter
    def assigned_class(self, new_class):
        """
        Setter method for the cfg_assigned_class attribute.
        """
        valid_classes = ["class3", "class4", "class6", "class8"]
        if new_class not in valid_classes:
            raise VerificationError(
                "Invalid class {0}, assigned class must be one of {1}".format(
                    new_class, ",".join(valid_classes)
                )
            )
        self.config["cfg_assigned_class"] = new_class

    @property
    @PyaoscxModule.materialized
    def priority(self):
        """
        Getter method for the priority attribute.

        :return: String value for priority.
        """
        return self.config["priority"]

    @priority.setter
    @PyaoscxModule.materialized
    def priority(self, new_priority):
        """
        Setter method for the priority attribute.
        """
        valid_priorities = ["low", "high", "critical"]
        if new_priority not in valid_priorities:
            raise VerificationError(
                "Invalid priority {0} priority must be one of {1}".format(
                    new_priority, ",".join(valid_priorities)
                )
            )
        self.config["priority"] = new_priority

    @property
    @PyaoscxModule.materialized
    def power_enabled(self):
        """
        Getter method for admin_disabled attribute.

        :return: Bool value for priority.
        """
        return not self.config["admin_disable"]

    @power_enabled.setter
    @PyaoscxModule.materialized
    def power_enabled(self, state):
        """
        Setter method for admin_disabled attribute
        """
        self.config["admin_disable"] = not state

    @property
    @PyaoscxModule.materialized
    def pd_class_override(self):
        """
        Getter method for pd_class_override attribute.

        :return: Bool value for pd_class_override.
        """
        return self.config["pd_class_override"]

    @pd_class_override.setter
    @PyaoscxModule.materialized
    def pd_class_override(self, new_flag):
        """
        Setter method for pd_class_override attribute
        """
        self.config["pd_class_override"] = new_flag

    @property
    @PyaoscxModule.materialized
    def pre_standard_detect(self):
        """
        Getter method for pre_standard_detect attribute.

        :return: Bool value for pre_standard_detect.
        """
        return self.config["pre_standard_detect"]

    @pre_standard_detect.setter
    @PyaoscxModule.materialized
    def pre_standard_detect(self, new_flag):
        """
        Setter method for pre_standard_detect attribute
        """
        self.config["pre_standard_detect"] = new_flag



# (C) Copyright 2019-2023 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

from netaddr import valid_mac

import json
import logging

from pyaoscx.exceptions.generic_op_error import GenericOperationError
from pyaoscx.exceptions.parameter_error import ParameterError
from pyaoscx.exceptions.response_error import ResponseError
from pyaoscx.exceptions.verification_error import VerificationError

from pyaoscx.utils import util as utils

from pyaoscx.pyaoscx_module import PyaoscxModule


class StaticMac(PyaoscxModule):
    """
    Provide configuration management for Static MAC on AOS-CX devices.
    """

    indices = ["mac_addr"]
    resource_uri_name = "static_macs"

    collection_uri = "system/vlans/{vlan_id}/static_macs"
    object_uri = collection_uri + "/{mac_addr}"

    def __init__(self, session, mac_addr, parent_vlan, uri=None, **kwargs):

        """
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device
        :param mac_addr: String MAC address, or netaddr EUI object.
            Example:
                '01:02:03:04:05:06'
        :param parent_vlan: Vlan object to which this MAC belongs
        :param uri: Optional string containing the uri of the MAC object
        """
        self.session = session
        # Assign id
        self.__mac_addr = mac_addr
        self.__parent_vlan = parent_vlan
        self._uri = uri
        # List used to determine attributes related to the Static MAC
        # configuration
        self.config_attrs = []
        self.materialized = False
        # Attribute dictionary used to manage the original data
        # obtained from the GET
        self._original_attributes = {}
        # Set arguments needed for correct creation
        utils.set_creation_attrs(self, **kwargs)
        # Attribute used to know if object was changed recently
        self.__modified = False
        uri_indices = {
            "vlan_id": self.__parent_vlan.id,
            "mac_addr": self.__mac_addr,
        }
        self._uri_indices = uri_indices
        self.base_uri = self.collection_uri.format(**uri_indices)
        self.path = self.object_uri.format(**uri_indices)

    @property
    def mac_addr(self):
        return self.__mac_addr

    @PyaoscxModule.connected
    def get(self, depth=None, selector=None):
        """
        Perform a GET call to retrieve data for a Static MAC table entry and
            fill the object with the incoming attributes

        :param depth: Integer deciding how many levels into the API JSON that
            references will be returned.
        :param selector: Alphanumeric option to select specific information to
            return.
        :return: Returns True if there is not an exception raised.
        """
        logging.info("Retrieving %s from switch", self)

        depth = depth or self.session.api.default_depth
        selector = selector or self.session.api.default_selector

        data = self._get_data(depth, selector)

        # Add dictionary as attributes for the object
        utils.create_attrs(self, data)

        # Determines if the Static MAC is configurable
        if selector in self.session.api.configurable_selectors:
            # Set self.config_attrs and delete ID from it
            utils.set_config_attrs(self, data, "config_attrs", self.indices)

        # Set original attributes
        self._original_attributes = data
        # Remove ID
        if "mac_addr" in self._original_attributes:
            self._original_attributes.pop("mac_addr")

        # Sets object as materialized
        # Information is loaded from the Device
        self.materialized = True

        return True

    @classmethod
    def get_all(cls, session, parent_vlan):
        """
        Perform a GET call to retrieve all system Static MACs inside a VLAN,
            and create a dictionary containing them.

        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param parent_vlan: VLAN object where Static MAC is stored.
        :return: Dictionary containing Static MAC IDs as keys and a Static MAC
            objects as values.
        """
        logging.info("Retrieving all %s data from switch", cls.__name__)

        uri = "{0}/{1}/static_macs".format(
            parent_vlan.base_uri, parent_vlan.id
        )

        try:
            response = session.request("GET", uri)
        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        data = json.loads(response.text)

        static_mac_dict = {}

        for uri in data.values():
            # Create a StaticMac object and adds it to parent Vrf object list
            mac_addr, static_mac = StaticMac.from_uri(
                session, parent_vlan, uri
            )
            # Load all Static MAC data from within the Switch
            static_mac_dict[mac_addr] = static_mac

        return static_mac_dict

    @PyaoscxModule.connected
    def apply(self):
        """
        Main method used to either create or update an existing Static MAC.
            Checks whether the Static MAC exists in the switch. Calls
            self.update() if Static MAC is being updated. Calls self.create()
            if a new Static MAC is being created.

        :return: Boolean, True if object was created or modified.
        """
        if not self.__parent_vlan.materialized:
            self.__parent_vlan.apply()
        if self.materialized:
            return self.update()
        return self.create()

    @PyaoscxModule.connected
    def update(self):
        """
        Perform a PUT call to apply changes to an existing Static MAC.

        :return: True if Object was modified and a PUT request was
            made.
        """
        static_mac_data = utils.get_attrs(self, self.config_attrs)

        if hasattr(self, "port") and self.port is not None:
            if isinstance(self.port, str):
                port = self.session.api.get_module(
                    self.session, "Interface", self.port
                )
            port.get()
            if not port.materialized:
                raise VerificationError(
                    "Port {0} not materialized".format(port.name)
                )
            if hasattr(port, "routing") and port.routing:
                raise VerificationError(
                    "{0} is not an L2 port".format(port.name)
                )
            static_mac_data["port"] = port.get_info_format()

        self.__modified = self._put_data(static_mac_data)
        return self.__modified

    @PyaoscxModule.connected
    def create(self):
        """
        Perform a POST call to create a new Static MAC using the object's
            attributes as POST body. Only returns if no exception is raised.

        :return: Boolean, True if entry was created
        """
        static_mac_data = utils.get_attrs(self, self.config_attrs)
        if not valid_mac(self.mac_addr):
            raise ParameterError("Invalid MAC Address")
        static_mac_data["mac_addr"] = self.mac_addr
        static_mac_data["vlan"] = self.__parent_vlan.get_info_format()
        if hasattr(self, "port") and self.port is not None:
            if isinstance(self.port, str):
                port = self.session.api.get_module(
                    self.session, "Interface", self.port
                )
            port.get()
            if not port.materialized:
                raise VerificationError(
                    "Port {0} not materialized".format(port.name)
                )
            if hasattr(port, "routing") and port.routing:
                raise VerificationError(
                    "{0} is not an L2 port".format(port.name)
                )
            static_mac_data["port"] = port.get_info_format()

        self.__modified = self._post_data(static_mac_data)
        return self.__modified

    @PyaoscxModule.connected
    def delete(self):
        """
        Perform DELETE call to delete Static MAC.
        """
        self._send_data(self.path, None, "DELETE", "Delete")

        # Delete back reference from VLAN
        for static_mac in self.__parent_vlan.static_macs:
            if static_mac.mac_address == self.mac_address:
                self.__parent_vlan.static_macs.remove(static_mac)

        utils.delete_attrs(self, self.config_attrs)

    @classmethod
    def from_response(cls, session, parent_vlan, response_data):
        """
        Create a Static MAC object given a response_data related to it.

        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param parent_vlan: parent Vlan object where Static MAC is stored.
        :param response_data: The response must be a dictionary of the form:
            {
            <mac_addr>: (
            "/rest/v10.04/system/vlans/<vlan_id>/static_macs/<mac_addr>"
            )
            }
        :return: Static MAC object
        """
        static_mac_arr = session.api.get_keys(
            response_data, StaticMac.resource_uri_name
        )
        mac_addr = static_mac_arr[0]
        return StaticMac(session, mac_addr, parent_vlan)

    @classmethod
    def from_uri(cls, session, parent_vlan, uri):
        """
        Create a StaticMac object given a URI.

        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param parent_vlan: VLAN object where Static MAC is stored.
        :param uri: a String with a URI.
        :return index, static_mac_obj: tuple containing both the StaticMac
            Object and the static_mac's mac_address.
        """
        # Obtain ID from URI
        # system/vlans/<vlan_id>/static_macs/<mac_addr>
        mac_addr = uri.split("/")[-1]

        # Create Static MAC object
        static_mac_obj = StaticMac(session, mac_addr, parent_vlan)

        return mac_addr, static_mac_obj

    def __str__(self):
        return "Static MAC ID {0}".format(self.mac_addr)

    @PyaoscxModule.deprecated
    def get_uri(self):
        """
        Method used to obtain the specific Statis MAC URI.

        :return: Object's URI.
        """
        return self.path

    @PyaoscxModule.deprecated
    def get_info_format(self):
        """
        Method used to obtain correct object format for referencing inside
            other objects.

        :return: Object format depending on the API Version.
        """
        return self.session.api.get_index(self)

    @property
    def modified(self):
        """
        Return boolean with whether this object has been modified.
        """
        return self.__modified

    @PyaoscxModule.deprecated
    def was_modified(self):
        """
        Getter method for the __modified attribute.

        :return: Boolean True if the object was recently modified.
        """
        return self.modified



# (C) Copyright 2022 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

import json
import logging
import re

from copy import deepcopy

from pyaoscx.exceptions.response_error import ResponseError
from pyaoscx.exceptions.generic_op_error import GenericOperationError
from pyaoscx.exceptions.verification_error import VerificationError

from pyaoscx.utils import util as utils

from pyaoscx.pyaoscx_module import PyaoscxModule


class Vni(PyaoscxModule):
    """
    Provide configuration management for Virtual Network IDs on
        AOS-CX devices
    """

    collection_uri = "system/virtual_network_ids"
    object_uri = collection_uri + "/{type},{id}"

    indices = ["type", "id"]
    resource_uri_name = "virtual_network_ids"

    def __init__(
        self, session, vni_id, interface, vni_type="vxlan_vni", **kwargs
    ):
        self.session = session
        self.id = vni_id
        self.type = vni_type
        self.config_attrs = []
        self.materialized = False
        self._original_attributes = {}
        self.interface = interface
        utils.create_attrs(self, kwargs)
        uri_indices = {
            "type": self.type,
            "id": self.id,
        }
        self.__modified = False
        self.materialized = False
        self.base_uri = self.collection_uri
        self.path = self.object_uri.format(**uri_indices)

    @PyaoscxModule.connected
    def get(self, depth=None, selector=None):
        """
        Perform a GET call to retrieve data for VNI table entry and fill
            the object with the incoming attributes.

        :param depth: Integer deciding how many levels into the API JSON that
            references will be returned
        :param selector: Alphanumeric option to select specific information to
            return.
        :return: Returns True if there is not an exception raised
        """
        logging.info("Retrieving %s from switch", self)
        # this is common for all PyaoscxModule derived classes
        self._get_and_copy_data(depth, selector, self.indices)
        selector = selector or self.session.api.default_selector
        data = self._get_data(depth, selector)

        for key, val in data.items():
            if key == "interface":
                continue
            setattr(self, key, val)

        Interface = self.session.api.get_module_class(
            self.session, "Interface"
        )

        intf = data["interface"]
        _, intf = Interface.from_uri(self.session, intf[next(iter(intf))])
        self.interface = intf

        self._original_attributes = deepcopy(data)

        if selector in self.session.api.configurable_selectors:
            setattr(
                self,
                "config_attrs",
                list(set(data.keys()) - set(self.indices)),
            )
        self.materialized = True

        if hasattr(self, "vrf") and self.vrf:
            Vrf = self.session.api.get_module_class(self.session, "Vrf")
            vrf_urls = self.vrf
            vrf_url = vrf_urls[next(iter(vrf_urls))]
            _, vrf = Vrf.from_uri(self.session, vrf_url)
            self.vrf = vrf
        if hasattr(self, "vlan") and self.vlan:
            Vlan = self.session.api.get_module_class(self.session, "Vlan")
            vlan_urls = self.vlan
            vlan_url = vlan_urls[next(iter(vlan_urls))]
            _, vlan = Vlan.from_uri(self.session, vlan_url)
            self.vlan = vlan
        return True

    @classmethod
    def get_all(cls, session, parent_intf):
        """
        Perform a GET call to retrieve all system VNI and create a dictionary
            containing each respective VNI

        :param cls: Object's class
        :param session: pyaoscx.Session object used to represent a logical
            connection to the switch
        :param parent_intf: Interface attached to this tunnel
        :return: Dictionary containing VNI IDs and a VNI object as value
        """
        logging.info("Retrieving all %s data from switch", cls.__name__)

        try:
            response = session.request("GET", cls.collection_uri)
        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        data = json.loads(response.text)

        vni_collection = {}
        for uri in data.values():
            vni_id, vni = cls.from_uri(session, parent_intf, uri)
            vni_collection[vni_id] = vni

        return vni_collection

    @PyaoscxModule.connected
    def apply(self):
        """
        Main method used to either create or update a
            VNI table entry

        :return: True if the object was modified
        """
        if self.materialized:
            return self.update()
        return self.create()

    @PyaoscxModule.connected
    def update(self):
        """
        Perform a PUT call to apply changes to an existing VNI table entry

        :return: True if object was modified and a PUT request was
            made. False otherwise
        """
        vni_data = utils.get_attrs(self, self.config_attrs)
        vni_data["interface"] = self.__interface.get_info_format()
        if hasattr(self, "routing"):
            vni_data["routing"] = self.routing
        if hasattr(self, "vrf") and self.vrf:
            vni_data["vrf"] = self.vrf.get_info_format()
        if hasattr(self, "vlan") and self.vlan:
            vni_data["vlan"] = self.vlan.get_info_format()
        self.__modified = self._put_data(vni_data)
        return self.__modified

    @PyaoscxModule.connected
    def create(self):
        """
        Perform a POST call to create a new VNI using the object's
            attributes as POST body. Exception is raised if object is unable
            to be created

        :return: Boolean, True if entry was created
        """
        vni_data = utils.get_attrs(self, self.config_attrs)
        vni_data["id"] = self.id
        vni_data["interface"] = self.__interface.get_info_format()
        vni_data["type"] = self.type
        if hasattr(self, "routing"):
            vni_data["routing"] = self.routing
        if hasattr(self, "vrf") and self.vrf:
            vni_data["vrf"] = self.vrf.get_info_format()
        if hasattr(self, "vlan") and self.vlan:
            vni_data["vlan"] = self.vlan.get_info_format()
        self.__modified = self._post_data(vni_data)
        return self.__modified

    @PyaoscxModule.connected
    def delete(self):
        """
        Perform a DELETE call to erase the VNI table entry.
        """
        self._send_data(self.path, None, "DELETE", "Delete")
        utils.delete_attrs(self, self.config_attrs)

    @classmethod
    def from_response(cls, session, parent_intf, response_data):
        """
        Create a VNI object given a response_data related to a VNI object

        :param cls: Object's class
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device
        :param parent_intf: Interface attached to this tunnel
        :param response_data: The response must be a dictionary of the form:
            {id: URL}
        :return: VNI object
        """
        uri = list(response_data.values())[0]

        return cls.from_uri(session, parent_intf, uri)

    @classmethod
    def from_uri(cls, session, parent_intf, uri):
        """
        Create a Vni object given a VNI URI

        :param cls: Object's class
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device
        :param parent_intf: Interface attached to this tunnel
        :param uri: A String with a URI
        :return: tuple containing the VNI ids (type, and actual
            ID) and the VNI object.
        """
        index_pattern = re.compile(
            r"(.*)virtual_network_ids/(?P<vni_type>.+),(?P<vni_id>.+)"
        )
        pattern_match = index_pattern.match(uri)
        vni_id = pattern_match.group("vni_id")
        vni_type = pattern_match.group("vni_type")
        vni_key = "{0},{1}".format(vni_type, vni_id)
        vni = cls(session, vni_id, parent_intf, vni_type=vni_type)
        return vni_key, vni

    def __str__(self):
        return "VNI {0}, type {1}".format(self.id, self.type)

    @PyaoscxModule.deprecated
    def get_uri(self):
        """
        Method used to obtain the specific VNI URI

        :return: Object's URI
        """
        uri = "{0}{1}".format(self.session.resource_prefix, self.path)
        return uri

    @PyaoscxModule.deprecated
    def get_info_format(self):
        """
        Method to obtain correct object format for referencing inside
            other objects

        :return: Object format depending on the API Version
        """
        return self.session.api.get_index(self)

    @property
    def modified(self):
        """
        Return boolean with whether this object has been modified.
        """
        return self.__modified

    @PyaoscxModule.deprecated
    def was_modified(self):
        """
        Getter method for the __modified attribute.

        :return: Boolean True if the object was recently modified.
        """
        return self.modified

    @property
    def interface(self):
        """
        Getter method for the interface attribute.

        :return: Interface object for method.
        """
        return self.__interface

    @interface.setter
    def interface(self, new_interface):
        """
        Setter method for the interface attribute.
        """
        if self.type == "vxlan_vni" and new_interface.type != "vxlan":
            raise VerificationError(
                "Incompatible VNI and Interface types: "
                "{0} and {1}".format(self.type, new_interface.type)
            )
        self.__interface = new_interface



# (C) Copyright 2019-2024 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

import os
import re

from ipaddress import ip_interface
from netaddr import mac_cisco, mac_unix_expanded
from netaddr import EUI as MacAddress
from netaddr.core import AddrFormatError
from requests_toolbelt.multipart.encoder import MultipartEncoder

from pyaoscx.exceptions.generic_op_error import GenericOperationError
from pyaoscx.exceptions.response_error import ResponseError
from pyaoscx.exceptions.parameter_error import ParameterError

ethertypes = {
    "aarp": 0x80F3,
    "appletalk": 0x809B,
    "arp": 0x0806,
    "fcoe": 0x8906,
    "fcoe-init": 0x8914,
    "ip": 0x0800,
    "ipv6": 0x86DD,
    "ipx-arpa": 0x8137,
    "ipx-non-arpa": 0x8138,
    "is-is": 0x22F4,
    "lldp": 0x88CC,
    "mpls-multicast": 0x8847,
    "mpls-unicast": 0x8848,
    "q-in-q": 0x9100,
    "rbridge": 0x8946,
    "trill": 0x22F3,
    "wake-on-lan": 0x0842,
}

r_ethertypes = {v: k for k, v in ethertypes.items()}

ip_protocols = {
    "ah": 51,
    "esp": 50,
    "gre": 47,
    "icmp": 1,
    "icmpv6": 58,
    "igmp": 2,
    "ospf": 89,
    "pim": 103,
    "sctp": 132,
    "tcp": 6,
    "udp": 17,
}

r_ip_protocols = {v: k for k, v in ip_protocols.items()}

dscp = {
    "AF11": 10,
    "AF12": 12,
    "AF13": 14,
    "AF21": 18,
    "AF22": 20,
    "AF23": 22,
    "AF31": 26,
    "AF32": 28,
    "AF33": 30,
    "AF41": 34,
    "AF42": 36,
    "AF43": 38,
    "CS0": 0,
    "CS1": 8,
    "CS2": 16,
    "CS3": 24,
    "CS4": 32,
    "CS5": 40,
    "CS6": 48,
    "CS7": 56,
    "EF": 46,
}

r_dscp = {v: k for k, v in dscp.items()}

icmp_types = {
    "echo-reply": 0,
    "destination-unreachable": 3,
    "source-quench": 4,
    "redirect": 5,
    "echo": 8,
    "router-advertisement": 9,
    "router-selection": 10,
    "time-exceeded": 11,
    "parameter-problem": 12,
    "timestamp": 13,
    "timestamp-reply": 14,
    "information-request": 15,
    "information-reply": 16,
    "address-mask-request": 17,
    "address-mask-reply": 18,
    "traceroute": 30,
    "extended-echo": 42,
    "extended-echo-reply": 43,
}

r_icmp_types = {v: k for k, v in icmp_types.items()}

icmpv6_types = {
    "destination-unreachable": 1,
    "packet-too-big": 2,
    "time-exceeded": 3,
    "parameter-problem": 4,
    "echo": 128,
    "echo-reply": 129,
    "multicast-listener-query": 130,
    "multicast-listener-report": 131,
    "multicast-listener-done": 132,
    "router-solicitation": 133,
    "router-advertisement": 134,
    "neighbor-solicitation": 135,
    "neighbor-advertisement": 136,
    "redirect-message": 137,
    "router-renumbering": 138,
    "icmp-node-information-query": 139,
    "icmp-node-information-response": 140,
    "mobile-prefix-solicitation": 146,
    "mobile-prefix-advertisement": 147,
    "duplicate-address-request-code-suffix": 157,
    "duplicate-address-confirmation-code-suffix": 158,
    "extended-echo": 160,
    "extended-echo-reply": 161,
}

r_icmpv6_types = {v: k for k, v in icmpv6_types.items()}

l4_ports = {
    "ftp-data": 20,
    "ftp": 21,
    "ssh": 22,
    "telnet": 23,
    "smtp": 25,
    "tacacs": 49,
    "dns": 53,
    "dhcp-server": 67,
    "dhcp-client": 68,
    "tftp": 69,
    "http": 80,
    "https": 443,
    "pop3": 110,
    "nntp": 119,
    "ntp": 123,
    "dce-rpc": 135,
    "netbios-ns": 137,
    "netbios-dgm": 138,
    "netbios-ssn": 139,
    "snmp": 161,
    "snmp-trap": 162,
    "bgp": 179,
    "ldap": 389,
    "microsoft-ds": 445,
    "isakmp": 500,
    "syslog": 514,
    "imap4": 585,
    "radius": 1812,
    "radius-acct": 1813,
    "iscsi": 3260,
    "rdp": 3389,
    "nat-t": 4500,
    "vxlan": 4789,
}

r_l4_ports = {v: k for k, v in l4_ports.items()}


def create_attrs(obj, data_dictionary):
    """
    Given a dictionary object creates class attributes. The methods implements
        setattr() which sets the value of the specified attribute of the
        specified object. If the attribute is already created within the
        object. It's state changes only if the current value is not None.
        Otherwise it keeps the previous value.
    :param data_dictionary: dictionary containing the attributes.
    """
    import copy

    # Used to create a deep copy of the dictionary
    dictionary_var = copy.deepcopy(data_dictionary)

    # K is the argument and V is the value of the given argument
    for k, v in dictionary_var.items():
        # In case a key has '-' inside it's name.
        k = k.replace("-", "_")
        obj.__dict__[k] = v


def get_dict_keys(dict):
    """
    Function used to get a list of all the keys of the respective dictionary.
    :param dict: Dictionary object used to obtain the keys.
    :return: List containing the keys of the given dictionary.
    """
    list = []
    for key in dict.keys():
        list.append(key)

    return list


def check_args(obj, **kwargs):
    """
    Given a object determines if the coming arguments are not already inside
        the object. If attribute is inside the config_attrs, it is ignored.
    :param obj: object in which the attributes are being set to.
    :param **kwargs list of arguments used to create the attributes.
    :return correct: True if all arguments are correct.
    """
    arguments = get_dict_keys(kwargs)
    correct = True
    for argument in arguments:
        if hasattr(obj, argument):
            correct = False
    return correct


def delete_attrs(obj, attr_list):
    """
    Given an object and a list of strings, delete attributes with the same name
        as the one inside the list.
    :param attr_list: List of attribute names that will be deleted from object
    """
    for attr in attr_list:
        if hasattr(obj, attr):
            delattr(obj, attr)


def get_attrs(obj, config_attrs):
    """
    Given an object obtains the attributes different to None.
    :param obj: object containing the attributes.
    :param config_attrs: a list of all the configurable attributes within the
        object.
    :return attr_data_dict: A dictionary containing all the attributes of the
        given object that have a value different to None.
    """
    attr_data_dict = {}
    for attr_name in config_attrs:
        attr_data_dict[attr_name] = getattr(obj, attr_name)
    return attr_data_dict


def set_creation_attrs(obj, **kwargs):
    """
    Used when instantiating the class with new attributes. Sets the
        configuration attributes list, for proper management of attributes
        related to configuration.
    :param obj: Python object in which attributes are being set.
    :param **kwargs: a dictionary containing the possible future arguments for
        the object.
    """
    if check_args(obj, **kwargs):
        obj.__dict__.update(kwargs)
        set_config_attrs(obj, kwargs)
    else:
        raise Exception(
            "ERROR: Trying to create existing attributes inside the object"
        )


def set_config_attrs(
    obj, config_dict, config_attrs="config_attrs", unwanted_attrs=[]
):
    """
    Add a list of strings inside the object to represent each attribute for
        config purposes.
    :param config_dict: Dictionary where each key represents an attribute.
    :param config_attrs: String containing the name of the attribute referring
        to a list.
    :param unwanted_attrs: Attributes that should be deleted, since they can't
        be modified.
    """
    # Set new configuration attributes list
    new_config_attrs = get_dict_keys(config_dict)

    # Delete unwanted attributes from configuration attributes list
    for element in unwanted_attrs:
        if element in new_config_attrs:
            # Remove all occurrences of element inside
            # the list representing the attributes related
            # to configuration
            new_config_attrs = list(filter((element).__ne__, new_config_attrs))
    # Set config attributes list with new values
    obj.__setattr__(config_attrs, new_config_attrs)


def _response_ok(response, call_type):
    """
    Checks whether API HTTP response contains the associated OK code.
    :param response: Response object.
    :param call_type: String containing the HTTP request type.
    :return: True if response was OK.
    """
    ok_codes = {
        "GET": [200],
        "PUT": [200, 204],
        "POST": [201],
        "DELETE": [204],
    }

    return response.status_code in ok_codes[call_type]


def file_upload(session, file_path, complete_uri, try_pycurl=False):
    """
    Upload any file given a URI and the path to a file located on the local
        machine.
    :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
    :param file_path: File name and path for local file uploading.
    :param complete_uri: Complete URI to perform the POST Request and upload
        the file. Example:
            "https://172.25.0.2/rest/v10.04/firmware?image=primary".
    :param try_pycurl: If True the function will try to use pycurl instead
        of requests.
    :return True if successful.
    """
    if try_pycurl:
        try:
            import pycurl
            from urllib.parse import urlencode

            use_pycurl = True
        except ImportError:
            use_pycurl = False
    else:
        use_pycurl = False

    file_name = os.path.basename(file_path)

    if use_pycurl:
        response = {}
        headers = {}

        def response_function(response_line):
            response["s"] = response_line.decode("iso-8859-1")
            return

        def header_function(header_line):
            header_line = header_line.decode("iso-8859-1")
            if ":" in header_line:
                name, value = header_line.split(":", 1)
                name = name.strip()
                value = value.strip()
                name = name.lower()
                if name == "set-cookie" and name in headers:
                    tmp_value = headers[name]
                    tmp_value = tmp_value.split(";")[0]
                    value = tmp_value + "; " + value
                headers[name] = value
            return

        # pycurl handles proxies diferently than requests
        for proto, ip in session.proxy.items():
            if proto in ["http", "https"]:
                proto = proto + "_proxy"
            if ip is None:
                os.environ.pop(proto, None)
            elif ip:
                os.environ[proto] = ip

        login_headers = ["Accept: */*", "x-use-csrf-token: true"]
        upload_headers = ["Accept: */*", "Content-Type: multipart/form-data"]
        logout_headers = ["Accept: */*"]
        login_data = {
            "username": session.username(),
            "password": session.password(),
        }
        postfields = urlencode(login_data)

        c = pycurl.Curl()
        c.setopt(c.URL, session.base_url + "login")
        c.setopt(c.HTTPHEADER, login_headers)
        c.setopt(c.CUSTOMREQUEST, "POST")
        c.setopt(c.POSTFIELDS, postfields)
        c.setopt(c.SSL_VERIFYPEER, False)
        c.setopt(c.SSL_VERIFYHOST, False)
        c.setopt(c.HEADERFUNCTION, header_function)
        c.setopt(c.WRITEFUNCTION, response_function)
        c.perform()
        c.close()

        token_id = "x-csrf-token"
        if "set-cookie" in headers:
            tmp_header = "Cookie: " + headers["set-cookie"]
            upload_headers.append(tmp_header)
            logout_headers.append(tmp_header)
        if token_id in headers:
            tmp_header = token_id + ": " + headers[token_id]
            upload_headers.append(tmp_header)
            logout_headers.append(tmp_header)

        c = pycurl.Curl()
        c.setopt(c.URL, complete_uri)
        c.setopt(c.HTTPHEADER, upload_headers)
        c.setopt(c.WRITEFUNCTION, response_function)
        c.setopt(c.CUSTOMREQUEST, "POST")
        c.setopt(c.SSL_VERIFYPEER, 0)
        c.setopt(c.SSL_VERIFYHOST, 0)
        c.setopt(
            c.HTTPPOST,
            [
                (
                    "fileupload",
                    (
                        c.FORM_FILE,
                        file_path,
                        c.FORM_FILENAME,
                        file_name,
                        c.FORM_CONTENTTYPE,
                        "application/octet-stream",
                    ),
                )
            ],
        )
        c.perform()
        status_code = c.getinfo(c.RESPONSE_CODE)
        c.close()

        c = pycurl.Curl()
        c.setopt(c.URL, session.base_url + "logout")
        c.setopt(c.HTTPHEADER, logout_headers)
        c.setopt(c.WRITEFUNCTION, response_function)
        c.setopt(c.CUSTOMREQUEST, "POST")
        c.setopt(c.SSL_VERIFYPEER, 0)
        c.setopt(c.SSL_VERIFYHOST, 0)
        c.perform()
        c.close()
        if status_code != 200:
            raise GenericOperationError(
                response["s"] + " (" + pycurl.version + ")", status_code
            )

        # Return true if successful
        return True
    # with requests
    else:
        file_param = {
            "fileupload": (
                file_name,
                open(file_path, "rb"),
                "application/octet-stream",
            )
        }
        m_part = MultipartEncoder(fields=file_param)
        file_header = {"Accept": "*/*", "Content-Type": m_part.content_type}

        try:
            file_header.update(session.s.headers)
            # Perform File Upload
            response_file_upload = session.s.post(
                complete_uri,
                verify=False,
                data=m_part,
                headers=file_header,
                proxies=session.proxy,
            )

        except Exception as e:
            raise ResponseError("POST", e)

        if response_file_upload.status_code != 200:
            raise GenericOperationError(
                response_file_upload.text, response_file_upload.status_code
            )

        # Return true if successful
        return True


def get_ip_version(ip):
    """
    Map if given IP address is v4 or v6. Will raise an exception if given
        address if invalid.
    :param ip: String with an IP address.
    :return: String with the IP version. Can be either ipv4 or ipv6.
    """
    try:
        if "/" in ip:
            ip_parts = iter(ip.split("/"))
            ip_addr = next(ip_parts)
            ip_mask = next(ip_parts)
            if not ip_mask.isnumeric():
                ip_mask = (
                    ipv6_netmask_to_cidr(ip_mask)
                    if ":" in ip_mask
                    else ipv4_netmask_to_cidr(ip_mask)
                )
                ip = ip_addr + "/" + str(ip_mask)
        ip_net = ip_interface(ip)
        return "ipv{0}".format(ip_net.version)
    except Exception as intr:
        msg = "Invalid IP: {0}".format(intr)
        raise ParameterError(msg)


def cidr_to_ipv4_netmask(cidr):
    """
    Convert CIDR mask format (/X) to net mask (/X.X.X.X)
    :param: CIDR mask to convert
    :return: String with IPv4 netmask
    """
    cidr = int(cidr)
    mask = (0xFFFFFFFF >> (32 - cidr)) << (32 - cidr)
    return (
        str((0xFF000000 & mask) >> 24)
        + "."
        + str((0x00FF0000 & mask) >> 16)
        + "."
        + str((0x0000FF00 & mask) >> 8)
        + "."
        + str((0x000000FF & mask))
    )


def ipv4_netmask_to_cidr(netmask):
    """
    Convert net mask (/X.X.X.X) to CIDR mask format (/X)
    :param: Net mask to convert
    :return: CIDR mask
    """
    return sum([bin(int(x)).count("1") for x in netmask.split(".")])


def cidr_to_ipv6_netmask(cidr):
    """
    Convert CIDR mask format (/X) to net mask (/X:X:X:X)
    :param: CIDR mask to convert
    :return: String with IPv6 netmask
    """
    cidr = int(cidr)
    all_bits = (2 << 127) - 1
    off_set = 128 - cidr
    raw_mask = hex(all_bits >> off_set << off_set)
    raw_mask = raw_mask.replace("0x", "").replace("L", "")
    mask_bytes = re.findall("....", raw_mask)
    new_mask = ":".join(mask_bytes)
    new_mask = new_mask.replace(":0000", "")
    if len(new_mask) < 39:
        new_mask += "::"
    return new_mask


def ipv6_netmask_to_cidr(netmask):
    """
    Convert net mask (/X:X:X:X) to CIDR mask format (/X)
    :param: Net mask to convert
    :return: CIDR mask
    """
    return sum(
        [
            bin(int("0x{0}".format(x if x != "" else "0"), 16)).count("1")
            for x in netmask.split(":")
        ]
    )


def fix_ip_mask(ip_address, version):
    """
    Fix Ip address mask in CIDR format to net mask
    :param ip_address: IP address with format (A.B.C.D/X)
    :param version: IP version (ipv4 or ipv6)
    :return: String with IP address with format (A.B.C.D/X.X.X.X)
    """
    if "/" in ip_address:
        ip_parts = iter(ip_address.split("/"))
        ip_addr = next(ip_parts)
        ip_mask = next(ip_parts)
        if ip_mask.isnumeric():
            ip_mask = (
                cidr_to_ipv4_netmask(ip_mask)
                if version == "ipv4"
                else cidr_to_ipv6_netmask(ip_mask)
            )
    else:  # host mask
        ip_addr = ip_address
        ip_mask = (
            "255.255.255.255"
            if version == "ipv4"
            else "ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff"
        )
    return ip_addr + "/" + ip_mask


def validate_mac_address(mac_addr, cisco_format=False):
    """
    Validate the correct formats for MAC address

    :param mac_addr: MAC address with any EUI format
    :param format: Flag to indicate if CISCO format is used
    :return: string with MAC address the format (XXXX.XXXX.XXXX)
        or (XX:XX:XX:XX:XX:XX)
    """
    mac_mask = ""
    if "/" in mac_addr:
        mac_addr = next(iter(mac_addr.split("/")))
    try:
        mac = MacAddress(mac_addr)
        if cisco_format:
            # ACL and Classifiers use CISCO format
            mac.dialect = mac_cisco
            # MAC addresses need a mask
            # that is the format required in REST
            mac_mask = "/ffff.ffff.ffff"
        else:
            # In general, REST resources use UNIX format
            mac.dialect = mac_unix_expanded
    except AddrFormatError as exc:
        raise ParameterError("Invalid MAC address: {0}".format(exc))

    return str(mac) + mac_mask


def set_acl(pyaoscx_module, acl_name, list_type, direction):
    """
    Attach ACL to an interface or vlan

    :param pyaoscx_module: Pyaoscx module (interface or vlan)
    :param acl_name: The name of the ACL.
    :param list_type: The type of the ACL (mac, ipv4 or ipv6).
    :param direction: The direction of the ACL (in, out, routed-in,
        routed-out)
    :return: True if the object was changed
    """
    Interface = pyaoscx_module.session.api.get_module_class(
        pyaoscx_module.session, "Interface"
    )
    Vlan = pyaoscx_module.session.api.get_module_class(
        pyaoscx_module.session, "Vlan"
    )
    is_interface = isinstance(pyaoscx_module, Interface)
    is_vlan = isinstance(pyaoscx_module, Vlan)

    if not is_interface and not is_vlan:
        raise ParameterError("Pyaoscx module must be Interface or Vlan")

    valid_types = ["mac", "ipv4", "ipv6"]
    valid_dirs = ["in", "out", "routed-in", "routed-out"]
    if list_type not in valid_types:
        raise ParameterError(
            "Invalid list_type {0}, valid types are: {1}".format(
                list_type, ", ".join(valid_types)
            )
        )
    if direction not in valid_dirs:
        raise ParameterError(
            "Invalid direction {0}, valid directions are {1}".format(
                direction, ", ".join(valid_dirs)
            )
        )
    is_l3 = False
    if is_interface:
        intf_type = (
            "port"
            if pyaoscx_module.type in [None, "lag"]
            else pyaoscx_module.type
        )
        if intf_type == "tunnel":
            intf_type = "tunnels"
        is_l3 = (
            hasattr(pyaoscx_module, "routing")
            and pyaoscx_module.routing
            or pyaoscx_module.type == "vlan"
        )
    else:
        intf_type = "vlan"
    gen_type = list_type.replace("ip", "")
    if is_interface and intf_type == "vlan" and "routed" not in direction:
        raise ParameterError(
            "Direction {0} not valid for VLAN Interfaces".format(direction)
        )
    # Create Acl object
    acl_obj = pyaoscx_module.session.api.get_module(
        pyaoscx_module.session, "ACL", index_id=acl_name, list_type=list_type
    )
    acl_obj.get()
    capability_prefix = "classifier_acl_{0}_".format(gen_type)

    # Verify direction capabilities
    needed_caps = []
    need_check_dir = False
    if direction == "in" and intf_type == "subinterface":
        suffix = "subinterface_in"
        needed_caps.append(capability_prefix + suffix)
        need_check_dir = True
    elif direction == "out":
        suffix = intf_type + "_out"
        needed_caps.append(capability_prefix + suffix)
        if is_l3:
            suffix = "routed_" + intf_type + "_out"
            needed_caps.append(capability_prefix + suffix)
        need_check_dir = True
    elif "routed" in direction:
        suffix = direction.replace("-", "_")
        needed_caps.append(capability_prefix + suffix)
        suffix = suffix.replace("routed", "routed_{0}".format(intf_type))
        needed_caps.append(capability_prefix + suffix)
        need_check_dir = True

    if (
        need_check_dir
        and [cap for cap in needed_caps if cap in acl_obj.capabilities] == []
    ):
        raise ParameterError(
            "{0}: ACL {1} {2} could not be applied".format(
                pyaoscx_module.name, list_type, direction
            )
        )
    gen_dir = "ingress" if direction in ["in", "routed-in"] else "egress"

    acl_attr = "acl{0}_{1}_cfg".format(gen_type, direction.replace("-", "_"))
    # Validate previous acl against acl object
    prev_acl = getattr(pyaoscx_module, acl_attr)
    if prev_acl == acl_obj:
        return False

    for ace in acl_obj.cfg_aces.values():
        if intf_type == "vlan" and hasattr(ace, "vlan") and ace.vlan:
            raise ParameterError(
                "{0}: VLAN ID cannot be used in an ACL applied to VLAN".format(
                    pyaoscx_module.name
                )
            )

        if hasattr(ace, "protocol"):
            # Verify special capabilities for protocols AH (51) and ESP (50)
            ah_cap = "classifier_ace_{0}_ah_{1}".format(gen_type, gen_dir)
            esp_cap = "classifier_ace_esp_egress"
            proto = ace.protocol
            if (
                proto == 51
                and (gen_type != "v6" or gen_dir == "egress")
                and ah_cap not in acl_obj.capabilities
            ):
                raise ParameterError(
                    "{0}: Protocol AH not supported for {1}".format(
                        pyaoscx_module.name, gen_dir
                    )
                )
            if (
                proto == 50
                and gen_dir == "egress"
                and esp_cap not in acl_obj.capabilities
            ):
                raise ParameterError(
                    "{0}: Protocol ESP not supported for {1}".format(
                        pyaoscx_module.name, gen_dir
                    )
                )
        # Verify capabilities for egress
        if gen_dir == "egress":
            if hasattr(ace, "fragment") and ace.fragment:
                cap_frg = "classifier_acl_{0}_frg_egress".format(gen_type)
                if (
                    "classifier_ace_frg_egress" not in acl_obj.capabilities
                    and cap_frg not in acl_obj.capabilities
                ):
                    raise ParameterError(
                        (
                            "{0}: {1} ACLS fragments"
                            " not supported for egress"
                        ).format(pyaoscx_module.name, acl_obj.list_type)
                    )
            if hasattr(ace, "log") and ace.log:
                cap_log = "classifier_acl_log_{0}_egress".format(ace.action)
                if cap_log not in acl_obj.capabilities:
                    raise ParameterError(
                        (
                            "{0}: Logging of ACL {1}"
                            " not supported for egress"
                        ).format(pyaoscx_module.name, ace.action)
                    )
            if (
                gen_type == "v4"
                and "classifier_ace_v4_tcp_flg_egress"
                not in acl_obj.capabilities
            ):
                for tcp_flag in ace.cap_tcp_flags:
                    if hasattr(ace, tcp_flag) and getattr(ace, tcp_flag):
                        raise ParameterError(
                            "{0}: TCP Flags not supported for egress".format(
                                pyaoscx_module.name
                            )
                        )

    # Finally set the ACL, also version number needs to be updated
    setattr(pyaoscx_module, acl_attr, acl_obj)
    attr_version = getattr(pyaoscx_module, acl_attr + "_version")
    setattr(
        pyaoscx_module,
        acl_attr + "_version",
        attr_version + 1 if attr_version else 1,
    )

    return pyaoscx_module.apply()


def clear_acl(pyaoscx_module, list_type, direction):
    """
    Removes ACL from an interface or vlan

    :param pyaoscx_module: Pyaoscx module (interface or vlan)
    :param list_type: The type of the ACL (mac, ipv4 or ipv6).
    :param direction: The direction of the ACL (in, out, routed-in,
        routed-out)
    :return: True if the object was changed
    """
    valid_types = ["mac", "ipv4", "ipv6"]
    valid_dirs = ["in", "out", "routed-in", "routed-out"]
    if list_type not in valid_types:
        raise ParameterError(
            "Invalid list_type {0}, valid types are: {1}".format(
                list_type, ", ".join(valid_types)
            )
        )
    if direction not in valid_dirs:
        raise ParameterError(
            "Invalid direction {0}, valid directions are {1}".format(
                direction, ", ".join(valid_dirs)
            )
        )
    acl_attr = "acl{0}_{1}_cfg".format(
        list_type.replace("ip", ""), direction.replace("-", "_")
    )

    if hasattr(pyaoscx_module, acl_attr):
        if getattr(pyaoscx_module, acl_attr) is None:
            return False
        setattr(pyaoscx_module, acl_attr, None)
        setattr(pyaoscx_module, acl_attr + "_version", 0)
    else:
        raise ParameterError(
            "{0}: ACL {1} {2} not supported in this platform".format(
                pyaoscx_module.name, list_type, direction
            )
        )

    return pyaoscx_module.apply()
